# -*- encoding: utf-8 -*-
#
# Copyright © 2012 eNovance <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import signal
import sys
import time

from ironic_lib.json_rpc import server as json_rpc
from oslo_config import cfg
from oslo_log import log
import oslo_messaging as messaging
from oslo_service import service
from oslo_utils import importutils

from ironic.common import context
from ironic.common import rpc
from ironic.objects import base as objects_base

LOG = log.getLogger(__name__)
CONF = cfg.CONF


class RPCService(service.Service):

    def __init__(self, host, manager_module, manager_class):
        super(RPCService, self).__init__()
        self.host = host
        manager_module = importutils.try_import(manager_module)
        manager_class = getattr(manager_module, manager_class)
        self.manager = manager_class(host, rpc.MANAGER_TOPIC)
        self.topic = self.manager.topic
        self.rpcserver = None
        self.deregister = True
        self._failure = None
        self._started = False

    def wait_for_start(self):
        while not self._started and not self._failure:
            time.sleep(0.1)
        if self._failure:
            LOG.critical(self._failure)
            sys.exit(self._failure)

    def start(self):
        self._failure = None
        self._started = False
        super(RPCService, self).start()
        try:
            self._real_start()
        except Exception as exc:
            self._failure = f"{exc.__class__.__name__}: {exc}"
            raise
        else:
            self._started = True

    def _real_start(self):
        admin_context = context.get_admin_context()

        serializer = objects_base.IronicObjectSerializer(is_server=True)
        # Perform preparatory actions before starting the RPC listener
        self.manager.prepare_host()
        if CONF.rpc_transport == 'json-rpc':
            self.rpcserver = json_rpc.WSGIService(
                self.manager, serializer, context.RequestContext.from_dict)
        elif CONF.rpc_transport != 'none':
            target = messaging.Target(topic=self.topic, server=self.host)
            endpoints = [self.manager]
            self.rpcserver = rpc.get_server(target, endpoints, serializer)

        if self.rpcserver is not None:
            self.rpcserver.start()

        self.handle_signal()
        self.manager.init_host(admin_context)
        rpc.set_global_manager(self.manager)

        LOG.info('Created RPC server with %(transport)s transport for service '
                 '%(service)s on host %(host)s.',
                 {'service': self.topic, 'host': self.host,
                  'transport': CONF.rpc_transport})

    def stop(self):
        try:
            if self.rpcserver is not None:
                self.rpcserver.stop()
                self.rpcserver.wait()
        except Exception as e:
            LOG.exception('Service error occurred when stopping the '
                          'RPC server. Error: %s', e)
        try:
            self.manager.del_host(deregister=self.deregister)
        except Exception as e:
            LOG.exception('Service error occurred when cleaning up '
                          'the RPC manager. Error: %s', e)

        super(RPCService, self).stop(graceful=True)
        LOG.info('Stopped RPC server for service %(service)s on host '
                 '%(host)s.',
                 {'service': self.topic, 'host': self.host})
        rpc.set_global_manager(None)

    def _handle_signal(self, signo, frame):
        LOG.info('Got signal SIGUSR1. Not deregistering on next shutdown '
                 'of service %(service)s on host %(host)s.',
                 {'service': self.topic, 'host': self.host})
        self.deregister = False

    def handle_signal(self):
        """Add a signal handler for SIGUSR1.

        The handler ensures that the manager is not deregistered when it is
        shutdown.
        """
        signal.signal(signal.SIGUSR1, self._handle_signal)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests to assert that audit middleware works as expected.
"""

from unittest import mock

from oslo_config import cfg
from oslo_middleware import healthcheck

from ironic.tests.unit.api import base


CONF = cfg.CONF


class TestHealthcheckMiddleware(base.BaseApiTest):
    """Provide a basic smoke test to ensure healthcheck middleware works."""

    @mock.patch.object(healthcheck, 'Healthcheck', autospec=True)
    def test_enable(self, mock_healthcheck):
        CONF.set_override('enabled', True, group='healthcheck')
        self._make_app()
        mock_healthcheck.assert_called_once_with(mock.ANY, CONF)

    @mock.patch.object(healthcheck, 'Healthcheck', autospec=True)
    def test_disable(self, mock_healthcheck):
        CONF.set_override('enabled', False, group='healthcheck')
        self._make_app()
        self.assertFalse(mock_healthcheck.called)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Test class for iBMC Management interface."""

import itertools
from unittest import mock

from oslo_utils import importutils

from ironic.common import boot_devices
from ironic.common import boot_modes
from ironic.common import exception
from ironic.conductor import task_manager
from ironic.drivers.modules.ibmc import mappings
from ironic.drivers.modules.ibmc import utils
from ironic.tests.unit.drivers.modules.ibmc import base

constants = importutils.try_import('ibmc_client.constants')
ibmc_client = importutils.try_import('ibmc_client')
ibmc_error = importutils.try_import('ibmc_client.exceptions')


class IBMCManagementTestCase(base.IBMCTestCase):

    def test_get_properties(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            properties = task.driver.get_properties()
            for prop in utils.COMMON_PROPERTIES:
                self.assertIn(prop, properties)

    @mock.patch.object(utils, 'parse_driver_info', autospec=True)
    def test_validate(self, mock_parse_driver_info):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.management.validate(task)
            mock_parse_driver_info.assert_called_once_with(task.node)

    def test_get_supported_boot_devices(self):
        if not mock._is_instance_mock(ibmc_client):
            mock.patch.object(ibmc_client, 'connect', autospec=True).start()
        connect_ibmc = ibmc_client.connect
        conn = self.mock_ibmc_conn(connect_ibmc)
        # mock return value
        _supported_boot_devices = list(mappings.GET_BOOT_DEVICE_MAP)
        conn.system.get.return_value = mock.Mock(
            boot_source_override=mock.Mock(
                supported_boot_devices=_supported_boot_devices
            )
        )
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            supported_boot_devices = (
                task.driver.management.get_supported_boot_devices(task))
            connect_ibmc.assert_called_with(**self.ibmc)
            expect = sorted(list(mappings.GET_BOOT_DEVICE_MAP.values()))
            self.assertEqual(expect, sorted(supported_boot_devices))

    def test_set_boot_device(self):
        if not mock._is_instance_mock(ibmc_client):
            mock.patch.object(ibmc_client, 'connect', autospec=True).start()
        connect_ibmc = ibmc_client.connect
        conn = self.mock_ibmc_conn(connect_ibmc)
        # mock return value
        conn.system.set_boot_source.return_value = None
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            device_mapping = [
                (boot_devices.PXE, constants.BOOT_SOURCE_TARGET_PXE),
                (boot_devices.DISK, constants.BOOT_SOURCE_TARGET_HDD),
                (boot_devices.CDROM, constants.BOOT_SOURCE_TARGET_CD),
                (boot_devices.BIOS,
                 constants.BOOT_SOURCE_TARGET_BIOS_SETUP),
                ('floppy', constants.BOOT_SOURCE_TARGET_FLOPPY),
            ]

            persistent_mapping = [
                (True, constants.BOOT_SOURCE_ENABLED_CONTINUOUS),
                (False, constants.BOOT_SOURCE_ENABLED_ONCE)
            ]

            data_source = list(itertools.product(device_mapping,
                                                 persistent_mapping))
            for (device, persistent) in data_source:
                task.driver.management.set_boot_device(
                    task, device[0], persistent=persistent[0])
                connect_ibmc.assert_called_with(**self.ibmc)
                conn.system.set_boot_source.assert_called_once_with(
                    device[1],
                    enabled=persistent[1])
                # Reset mocks
                connect_ibmc.reset_mock()
                conn.system.set_boot_source.reset_mock()

    def test_set_boot_device_fail(self):
        if not mock._is_instance_mock(ibmc_client):
            mock.patch.object(ibmc_client, 'connect', autospec=True).start()
        connect_ibmc = ibmc_client.connect
        conn = self.mock_ibmc_conn(connect_ibmc)
        # mock return value
        conn.system.set_boot_source.side_effect = (
            ibmc_error.IBMCClientError
        )

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.assertRaisesRegex(
                exception.IBMCError, 'set iBMC boot device',
                task.driver.management.set_boot_device, task,
                boot_devices.PXE)
            connect_ibmc.assert_called_with(**self.ibmc)
            conn.system.set_boot_source.assert_called_once_with(
                constants.BOOT_SOURCE_TARGET_PXE,
                enabled=constants.BOOT_SOURCE_ENABLED_ONCE)

    def test_get_boot_device(self):
        if not mock._is_instance_mock(ibmc_client):
            mock.patch.object(ibmc_client, 'connect', autospec=True).start()
        connect_ibmc = ibmc_client.connect
        conn = self.mock_ibmc_conn(connect_ibmc)
        # mock return value
        conn.system.get.return_value = mock.Mock(
            boot_source_override=mock.Mock(
                target=constants.BOOT_SOURCE_TARGET_PXE,
                enabled=constants.BOOT_SOURCE_ENABLED_CONTINUOUS
            )
        )
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            result_boot_device = task.driver.management.get_boot_device(task)
            conn.system.get.assert_called_once()
            connect_ibmc.assert_called_once_with(**self.ibmc)
            expected = {'boot_device': boot_devices.PXE,
                        'persistent': True}
            self.assertEqual(expected, result_boot_device)

    def test_get_supported_boot_modes(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            supported_boot_modes = (
                task.driver.management.get_supported_boot_modes(task))
            self.assertEqual(list(mappings.SET_BOOT_MODE_MAP),
                             supported_boot_modes)

    def test_set_boot_mode(self):
        if not mock._is_instance_mock(ibmc_client):
            mock.patch.object(ibmc_client, 'connect', autospec=True).start()
        connect_ibmc = ibmc_client.connect
        conn = self.mock_ibmc_conn(connect_ibmc)
        # mock system boot source override return value
        conn.system.get.return_value = mock.Mock(
            boot_source_override=mock.Mock(
                target=constants.BOOT_SOURCE_TARGET_PXE,
                enabled=constants.BOOT_SOURCE_ENABLED_CONTINUOUS
            )
        )
        conn.system.set_boot_source.return_value = None

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            expected_values = [
                (boot_modes.LEGACY_BIOS, constants.BOOT_SOURCE_MODE_BIOS),
                (boot_modes.UEFI, constants.BOOT_SOURCE_MODE_UEFI)
            ]

            for ironic_boot_mode, ibmc_boot_mode in expected_values:
                task.driver.management.set_boot_mode(task,
                                                     mode=ironic_boot_mode)

                conn.system.get.assert_called_once()
                connect_ibmc.assert_called_with(**self.ibmc)

                conn.system.set_boot_source.assert_called_once_with(
                    constants.BOOT_SOURCE_TARGET_PXE,
                    enabled=constants.BOOT_SOURCE_ENABLED_CONTINUOUS,
                    mode=ibmc_boot_mode)

                # Reset
                connect_ibmc.reset_mock()
                conn.system.set_boot_source.reset_mock()
                conn.system.get.reset_mock()

    def test_set_boot_mode_fail(self):
        if not mock._is_instance_mock(ibmc_client):
            mock.patch.object(ibmc_client, 'connect', autospec=True).start()
        connect_ibmc = ibmc_client.connect
        conn = self.mock_ibmc_conn(connect_ibmc)
        # mock system boot source override return value
        conn.system.get.return_value = mock.Mock(
            boot_source_override=mock.Mock(
                target=constants.BOOT_SOURCE_TARGET_PXE,
                enabled=constants.BOOT_SOURCE_ENABLED_CONTINUOUS
            )
        )
        conn.system.set_boot_source.side_effect = (
            ibmc_error.IBMCClientError
        )

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            expected_values = [
                (boot_modes.LEGACY_BIOS, constants.BOOT_SOURCE_MODE_BIOS),
                (boot_modes.UEFI, constants.BOOT_SOURCE_MODE_UEFI)
            ]

            for ironic_boot_mode, ibmc_boot_mode in expected_values:
                self.assertRaisesRegex(
                    exception.IBMCError, 'set iBMC boot mode',
                    task.driver.management.set_boot_mode, task,
                    ironic_boot_mode)

                conn.system.set_boot_source.assert_called_once_with(
                    constants.BOOT_SOURCE_TARGET_PXE,
                    enabled=constants.BOOT_SOURCE_ENABLED_CONTINUOUS,
                    mode=ibmc_boot_mode)

                conn.system.get.assert_called_once()
                connect_ibmc.assert_called_with(**self.ibmc)

                # Reset
                connect_ibmc.reset_mock()
                conn.system.set_boot_source.reset_mock()
                conn.system.get.reset_mock()

    def test_get_boot_mode(self):
        if not mock._is_instance_mock(ibmc_client):
            mock.patch.object(ibmc_client, 'connect', autospec=True).start()
        connect_ibmc = ibmc_client.connect
        conn = self.mock_ibmc_conn(connect_ibmc)
        # mock system boot source override return value
        conn.system.get.return_value = mock.Mock(
            boot_source_override=mock.Mock(
                target=constants.BOOT_SOURCE_TARGET_PXE,
                enabled=constants.BOOT_SOURCE_ENABLED_CONTINUOUS,
                mode=constants.BOOT_SOURCE_MODE_BIOS,
            )
        )
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            response = task.driver.management.get_boot_mode(task)

            conn.system.get.assert_called_once()
            connect_ibmc.assert_called_with(**self.ibmc)

            expected = boot_modes.LEGACY_BIOS
            self.assertEqual(expected, response)

    def test_get_sensors_data(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaises(NotImplementedError,
                              task.driver.management.get_sensors_data, task)

    def test_inject_nmi(self):
        if not mock._is_instance_mock(ibmc_client):
            mock.patch.object(ibmc_client, 'connect', autospec=True).start()
        connect_ibmc = ibmc_client.connect
        conn = self.mock_ibmc_conn(connect_ibmc)
        # mock system boot source override return value
        conn.system.reset.return_value = None
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.inject_nmi(task)

            connect_ibmc.assert_called_with(**self.ibmc)
            conn.system.reset.assert_called_once_with(constants.RESET_NMI)

    def test_inject_nmi_fail(self):
        if not mock._is_instance_mock(ibmc_client):
            mock.patch.object(ibmc_client, 'connect', autospec=True).start()
        connect_ibmc = ibmc_client.connect
        conn = self.mock_ibmc_conn(connect_ibmc)
        # mock system boot source override return value
        conn.system.reset.side_effect = (
            ibmc_error.IBMCClientError
        )
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.assertRaisesRegex(
                exception.IBMCError, 'inject iBMC NMI',
                task.driver.management.inject_nmi, task)

            connect_ibmc.assert_called_with(**self.ibmc)
            conn.system.reset.assert_called_once_with(constants.RESET_NMI)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add boot_mode and secure_boot

Revision ID: c1846a214450
Revises: 2bbd96b6ccb9
Create Date: 2021-06-21 15:57:37.330442

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'c1846a214450'
down_revision = '2bbd96b6ccb9'


def upgrade():
    op.add_column('nodes', sa.Column('boot_mode',
                  sa.String(length=16), nullable=True))
    op.add_column('nodes', sa.Column('secure_boot',
                  sa.Boolean(), nullable=True))



# Copyright 2018 DMTF. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import importutils

from ironic.common import exception
from ironic.common import states
from ironic.conductor import task_manager
from ironic.conductor import utils as manager_utils
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules.redfish import bios as redfish_bios
from ironic.drivers.modules.redfish import boot as redfish_boot
from ironic.drivers.modules.redfish import utils as redfish_utils
from ironic import objects
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.db import utils as db_utils
from ironic.tests.unit.objects import utils as obj_utils

sushy = importutils.try_import('sushy')

INFO_DICT = db_utils.get_test_redfish_info()


class NoBiosSystem(object):
    identity = '/redfish/v1/Systems/1234'

    @property
    def bios(self):
        raise sushy.exceptions.MissingAttributeError(attribute='Bios',
                                                     resource=self)


@mock.patch('oslo_utils.eventletutils.EventletEvent.wait',
            lambda *args, **kwargs: None)
class RedfishBiosTestCase(db_base.DbTestCase):

    def setUp(self):
        super(RedfishBiosTestCase, self).setUp()
        self.config(enabled_bios_interfaces=['redfish'],
                    enabled_hardware_types=['redfish'],
                    enabled_power_interfaces=['redfish'],
                    enabled_boot_interfaces=['redfish-virtual-media'],
                    enabled_management_interfaces=['redfish'])
        self.node = obj_utils.create_test_node(
            self.context, driver='redfish', driver_info=INFO_DICT)

    @mock.patch.object(redfish_bios, 'sushy', None)
    def test_loading_error(self):
        self.assertRaisesRegex(
            exception.DriverLoadError,
            'Unable to import the sushy library',
            redfish_bios.RedfishBIOS)

    def test_get_properties(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            properties = task.driver.get_properties()
            for prop in redfish_utils.COMMON_PROPERTIES:
                self.assertIn(prop, properties)

    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    def test_validate(self, mock_parse_driver_info):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.bios.validate(task)
            mock_parse_driver_info.assert_called_once_with(task.node)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    @mock.patch.object(objects, 'BIOSSettingList', autospec=True)
    def test_cache_bios_settings_noop(self, mock_setting_list,
                                      mock_get_system):
        create_list = []
        update_list = []
        delete_list = []
        nochange_list = [{'name': 'EmbeddedSata', 'value': 'Raid'},
                         {'name': 'NicBoot1', 'value': 'NetworkBoot'}]
        mock_setting_list.sync_node_setting.return_value = (
            create_list, update_list, delete_list, nochange_list
        )

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            settings = {'foo': 'bar'}
            mock_get_system.return_value.bios.attributes = settings

            task.driver.bios.cache_bios_settings(task)
            mock_get_system.assert_called_once_with(task.node)
            mock_setting_list.sync_node_setting.assert_called_once_with(
                task.context, task.node.id,
                [{'name': 'foo', 'value': 'bar'}])

            mock_setting_list.create.assert_not_called()
            mock_setting_list.save.assert_not_called()
            mock_setting_list.delete.assert_not_called()

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    @mock.patch.object(objects, 'BIOSSettingList', autospec=True)
    def test_cache_bios_settings_no_bios(self, mock_setting_list,
                                         mock_get_system):
        create_list = []
        update_list = []
        delete_list = []
        nochange_list = [{'name': 'EmbeddedSata', 'value': 'Raid'},
                         {'name': 'NicBoot1', 'value': 'NetworkBoot'}]
        mock_setting_list.sync_node_setting.return_value = (
            create_list, update_list, delete_list, nochange_list
        )
        mock_get_system.return_value = NoBiosSystem()

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaisesRegex(exception.UnsupportedDriverExtension,
                                   'BIOS settings are not supported',
                                   task.driver.bios.cache_bios_settings, task)
            mock_get_system.assert_called_once_with(task.node)
            mock_setting_list.sync_node_setting.assert_not_called()
            mock_setting_list.create.assert_not_called()
            mock_setting_list.save.assert_not_called()
            mock_setting_list.delete.assert_not_called()

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    @mock.patch.object(objects, 'BIOSSettingList', autospec=True)
    def test_cache_bios_settings(self, mock_setting_list, mock_get_system):
        create_list = [{'name': 'DebugMode', 'value': 'enabled'}]
        update_list = [{'name': 'BootMode', 'value': 'Uefi'},
                       {'name': 'NicBoot2', 'value': 'NetworkBoot'}]
        delete_list = [{'name': 'AdminPhone', 'value': '555-867-5309'}]
        nochange_list = [{'name': 'EmbeddedSata', 'value': 'Raid'},
                         {'name': 'NicBoot1', 'value': 'NetworkBoot'}]
        delete_names = []
        for setting in delete_list:
            delete_names.append(setting.get('name'))
        mock_setting_list.sync_node_setting.return_value = (
            create_list, update_list, delete_list, nochange_list
        )

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            settings = {'foo': 'bar'}
            mock_get_system.return_value.bios.attributes = settings

            task.driver.bios.cache_bios_settings(task)
            mock_get_system.assert_called_once_with(task.node)
            mock_setting_list.sync_node_setting.assert_called_once_with(
                task.context, task.node.id,
                [{'name': 'foo', 'value': 'bar'}])

            mock_setting_list.create.assert_called_once_with(
                task.context, task.node.id, create_list)
            mock_setting_list.save.assert_called_once_with(
                task.context, task.node.id, update_list)
            mock_setting_list.delete.assert_called_once_with(
                task.context, task.node.id, delete_names)

    @mock.patch.object(redfish_boot.RedfishVirtualMediaBoot, 'prepare_ramdisk',
                       spec_set=True, autospec=True)
    @mock.patch.object(deploy_utils, 'build_agent_options', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    @mock.patch.object(manager_utils, 'node_power_action', autospec=True)
    def _test_step_pre_reboot(self, mock_power_action, mock_get_system,
                              mock_build_agent_options, mock_prepare):
        if self.node.clean_step:
            step_data = self.node.clean_step
            check_fields = ['cleaning_reboot', 'skip_current_clean_step']
            expected_ret = states.CLEANWAIT
        else:
            step_data = self.node.deploy_step
            check_fields = ['deployment_reboot', 'skip_current_deploy_step']
            expected_ret = states.DEPLOYWAIT
        data = step_data['argsinfo'].get('settings', None)
        step = step_data['step']
        if step == 'factory_reset':
            check_fields.append('post_factory_reset_reboot_requested')
        elif step == 'apply_configuration':
            check_fields.append('post_config_reboot_requested')
            attributes = {s['name']: s['value'] for s in data}
        mock_build_agent_options.return_value = {'a': 'b'}
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            bios = mock_get_system(task.node).bios
            if step == 'factory_reset':
                ret = task.driver.bios.factory_reset(task)
            if step == 'apply_configuration':
                bios.supported_apply_times = []
                ret = task.driver.bios.apply_configuration(task, data)
            mock_get_system.assert_called_with(task.node)
            mock_power_action.assert_called_once_with(task, states.REBOOT)
            if step == 'factory_reset':
                bios.reset_bios.assert_called_once()
            if step == 'apply_configuration':
                bios.set_attributes.assert_called_once_with(
                    attributes, apply_time=None)
            mock_build_agent_options.assert_called_once_with(task.node)
            mock_prepare.assert_called_once_with(mock.ANY, task, {'a': 'b'})
            info = task.node.driver_internal_info
            self.assertTrue(all(x in info for x in check_fields))
            self.assertEqual(expected_ret, ret)

    def test_factory_reset_step_pre_reboot_cleaning(self):
        self.node.clean_step = {'priority': 100, 'interface': 'bios',
                                'step': 'factory_reset', 'argsinfo': {}}
        self.node.save()
        self._test_step_pre_reboot()

    def test_factory_reset_step_pre_reboot_deploying(self):
        self.node.deploy_step = {'priority': 100, 'interface': 'bios',
                                 'step': 'factory_reset', 'argsinfo': {}}
        self.node.save()
        self._test_step_pre_reboot()

    def test_apply_conf_step_pre_reboot_cleaning(self):
        data = [{'name': 'ProcTurboMode', 'value': 'Disabled'},
                {'name': 'NicBoot1', 'value': 'NetworkBoot'}]
        self.node.clean_step = {'priority': 100, 'interface': 'bios',
                                'step': 'apply_configuration',
                                'argsinfo': {'settings': data}}
        self.node.save()
        self._test_step_pre_reboot()

    def test_apply_conf_step_pre_reboot_deploying(self):
        data = [{'name': 'ProcTurboMode', 'value': 'Disabled'},
                {'name': 'NicBoot1', 'value': 'NetworkBoot'}]
        self.node.deploy_step = {'priority': 100, 'interface': 'bios',
                                 'step': 'apply_configuration',
                                 'argsinfo': {'settings': data}}
        self.node.save()
        self._test_step_pre_reboot()

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def _test_step_post_reboot(self, mock_get_system,
                               attributes_after_reboot=None):
        if self.node.deploy_step:
            step_data = self.node.deploy_step
        else:
            step_data = self.node.clean_step
        data = step_data['argsinfo'].get('settings', None)
        step = step_data['step']
        if step == 'factory_reset':
            check_fields = ['post_factory_reset_reboot_requested']
        if step == 'apply_configuration':
            mock_bios = mock.Mock()
            # if attributes after reboot not provided then mimic success
            # by returning the same as requested
            mock_bios.attributes = attributes_after_reboot \
                or self.node.driver_internal_info['requested_bios_attrs']
            mock_bios.get_attribute_registry = []
            mock_system = mock.Mock()
            mock_system.bios = mock_bios
            mock_get_system.return_value = mock_system
            check_fields = ['post_config_reboot_requested',
                            'requested_bios_attrs']
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            if step == 'factory_reset':
                task.driver.bios.factory_reset(task)
            if step == 'apply_configuration':
                task.driver.bios.apply_configuration(task, data)
            mock_get_system.assert_called_with(task.node)
            info = task.node.driver_internal_info
            for field in check_fields:
                self.assertNotIn(field, info)

    def test_factory_reset_post_reboot_cleaning(self):
        self.node.clean_step = {'priority': 100, 'interface': 'bios',
                                'step': 'factory_reset', 'argsinfo': {}}
        node = self.node
        driver_internal_info = node.driver_internal_info
        driver_internal_info['post_factory_reset_reboot_requested'] = True
        node.driver_internal_info = driver_internal_info
        node.save()
        self._test_step_post_reboot()

    def test_factory_reset_post_reboot_deploying(self):
        self.node.deploy_step = {'priority': 100, 'interface': 'bios',
                                 'step': 'factory_reset', 'argsinfo': {}}
        node = self.node
        driver_internal_info = node.driver_internal_info
        driver_internal_info['post_factory_reset_reboot_requested'] = True
        node.driver_internal_info = driver_internal_info
        node.save()
        self._test_step_post_reboot()

    @mock.patch.object(manager_utils, 'cleaning_error_handler', autospec=True)
    @mock.patch.object(manager_utils, 'deploying_error_handler', autospec=True)
    def test_apply_conf_post_reboot_cleaning(self,
                                             mock_deploying_error_handler,
                                             mock_cleaning_error_handler):
        data = [{'name': 'ProcTurboMode', 'value': 'Enabled'},
                {'name': 'NicBoot1', 'value': 'NetworkBoot'}]
        self.node.clean_step = {'priority': 100, 'interface': 'bios',
                                'step': 'apply_configuration',
                                'argsinfo': {'settings': data}}
        requested_attrs = {'ProcTurboMode': 'Enabled',
                           'NicBoot1': 'NetworkBoot'}
        node = self.node
        driver_internal_info = node.driver_internal_info
        driver_internal_info['post_config_reboot_requested'] = True
        driver_internal_info['requested_bios_attrs'] = requested_attrs
        self.node.driver_internal_info = driver_internal_info
        self.node.save()
        self._test_step_post_reboot()
        mock_cleaning_error_handler.assert_not_called()
        mock_deploying_error_handler.assert_not_called()

    @mock.patch.object(manager_utils, 'cleaning_error_handler', autospec=True)
    def test_apply_conf_post_reboot_cleaning_failed(
            self, mock_cleaning_error_handler):
        data = [{'name': 'ProcTurboMode', 'value': 'Enabled',
                 'registry': {'description': 'Turbo mode'}}]
        self.node.clean_step = {'priority': 100, 'interface': 'bios',
                                'step': 'apply_configuration',
                                'argsinfo': {'settings': data}}
        requested_attrs = {'ProcTurboMode': 'Enabled'}
        attributes_after_reboot = {'ProcTurboMode': 'Disabled'}
        node = self.node
        driver_internal_info = node.driver_internal_info
        driver_internal_info['post_config_reboot_requested'] = True
        driver_internal_info['requested_bios_attrs'] = requested_attrs
        self.node.driver_internal_info = driver_internal_info
        self.node.provision_state = states.CLEANING
        self.node.save()
        self._test_step_post_reboot(
            attributes_after_reboot=attributes_after_reboot)
        mock_cleaning_error_handler.assert_called_once()

    def test_apply_conf_post_reboot_deploying(self):
        data = [{'name': 'ProcTurboMode', 'value': 'Enabled',
                 'registry': {'description': 'Turbo mode'}},
                {'name': 'NicBoot1', 'value': 'NetworkBoot',
                 'registry': {'description': 'Boot off network'}}]
        self.node.deploy_step = {'priority': 100, 'interface': 'bios',
                                 'step': 'apply_configuration',
                                 'argsinfo': {'settings': data}}
        requested_attrs = {'ProcTurboMode': 'Enabled',
                           'NicBoot1': 'NetworkBoot'}
        node = self.node
        driver_internal_info = node.driver_internal_info
        driver_internal_info['post_config_reboot_requested'] = True
        driver_internal_info['requested_bios_attrs'] = requested_attrs
        self.node.driver_internal_info = driver_internal_info
        self.node.save()
        self._test_step_post_reboot()

    @mock.patch.object(manager_utils, 'deploying_error_handler', autospec=True)
    def test_apply_conf_post_reboot_deploying_failed(
            self, mock_deploying_error_handler):
        data = [{'name': 'ProcTurboMode', 'value': 'Enabled',
                 'registry': {'description': 'Turbo mode'}}]
        self.node.deploy_step = {'priority': 100, 'interface': 'bios',
                                 'step': 'apply_configuration',
                                 'argsinfo': {'settings': data}}
        requested_attrs = {'ProcTurboMode': 'Enabled'}
        attributes_after_reboot = {'ProcTurboMode': 'Disabled'}
        node = self.node
        driver_internal_info = node.driver_internal_info
        driver_internal_info['post_config_reboot_requested'] = True
        driver_internal_info['requested_bios_attrs'] = requested_attrs
        self.node.driver_internal_info = driver_internal_info
        self.node.provision_state = states.DEPLOYWAIT
        self.node.save()
        self._test_step_post_reboot(
            attributes_after_reboot=attributes_after_reboot)
        mock_deploying_error_handler.assert_called_once()

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_factory_reset_fail(self, mock_get_system):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            bios = mock_get_system(task.node).bios
            bios.reset_bios.side_effect = sushy.exceptions.SushyError
            self.assertRaisesRegex(
                exception.RedfishError, 'BIOS factory reset failed',
                task.driver.bios.factory_reset, task)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_factory_reset_not_supported(self, mock_get_system):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            mock_get_system.return_value = NoBiosSystem()
            self.assertRaisesRegex(
                exception.RedfishError, 'BIOS factory reset failed',
                task.driver.bios.factory_reset, task)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_apply_configuration_not_supported(self, mock_get_system):
        settings = [{'name': 'ProcTurboMode', 'value': 'Disabled'},
                    {'name': 'NicBoot1', 'value': 'NetworkBoot'}]
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            mock_get_system.return_value = NoBiosSystem()
            self.assertRaisesRegex(exception.RedfishError,
                                   'BIOS settings are not supported',
                                   task.driver.bios.apply_configuration,
                                   task, settings)
            mock_get_system.assert_called_once_with(task.node)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_check_bios_attrs(self, mock_get_system):
        settings = [{'name': 'ProcTurboMode', 'value': 'Disabled'},
                    {'name': 'NicBoot1', 'value': 'NetworkBoot'}]
        requested_attrs = {'ProcTurboMode': 'Enabled',
                           'NicBoot1': 'NetworkBoot'}
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            attributes = mock_get_system(task.node).bios.attributes
            task.node.driver_internal_info[
                'post_config_reboot_requested'] = True
            task.node.driver_internal_info[
                'requested_bios_attrs'] = requested_attrs
            task.driver.bios._check_bios_attrs = mock.MagicMock()
            task.driver.bios.apply_configuration(task, settings)
            task.driver.bios._check_bios_attrs \
                .assert_called_once_with(task, attributes, requested_attrs)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_apply_configuration_fail(self, mock_get_system):
        settings = [{'name': 'ProcTurboMode', 'value': 'Disabled'},
                    {'name': 'NicBoot1', 'value': 'NetworkBoot'}]
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            bios = mock_get_system(task.node).bios
            bios.set_attributes.side_effect = sushy.exceptions.SushyError
            self.assertRaisesRegex(
                exception.RedfishError, 'BIOS apply configuration failed',
                task.driver.bios.apply_configuration, task, settings)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_post_configuration(self, mock_get_system):
        settings = [{'name': 'ProcTurboMode', 'value': 'Disabled'},
                    {'name': 'NicBoot1', 'value': 'NetworkBoot'}]
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.bios.post_configuration = mock.MagicMock()
            task.driver.bios.apply_configuration(task, settings)
            task.driver.bios.post_configuration\
                .assert_called_once_with(task, settings)

    @mock.patch.object(redfish_boot.RedfishVirtualMediaBoot, 'prepare_ramdisk',
                       spec_set=True, autospec=True)
    @mock.patch.object(deploy_utils, 'build_agent_options', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    @mock.patch.object(manager_utils, 'node_power_action', autospec=True)
    def test_apply_configuration_apply_time_immediate(self, mock_power_action,
                                                      mock_get_system,
                                                      mock_build_agent_options,
                                                      mock_prepare):
        settings = [{'name': 'ProcTurboMode', 'value': 'Disabled'},
                    {'name': 'NicBoot1', 'value': 'NetworkBoot'}]
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            bios = mock_get_system(task.node).bios
            bios.supported_apply_times = ['immediate']

            task.driver.bios.apply_configuration(task, settings)

            bios.set_attributes.assert_called_once_with(
                {s['name']: s['value'] for s in settings},
                apply_time=None)

    @mock.patch.object(redfish_boot.RedfishVirtualMediaBoot, 'prepare_ramdisk',
                       spec_set=True, autospec=True)
    @mock.patch.object(deploy_utils, 'build_agent_options', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    @mock.patch.object(manager_utils, 'node_power_action', autospec=True)
    def test_apply_configuration_apply_time_on_reset(self, mock_power_action,
                                                     mock_get_system,
                                                     mock_build_agent_options,
                                                     mock_prepare):
        settings = [{'name': 'ProcTurboMode', 'value': 'Disabled'},
                    {'name': 'NicBoot1', 'value': 'NetworkBoot'}]
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            bios = mock_get_system(task.node).bios
            bios.supported_apply_times = [sushy.APPLY_TIME_ON_RESET]

            task.driver.bios.apply_configuration(task, settings)

            bios.set_attributes.assert_called_once_with(
                {s['name']: s['value'] for s in settings},
                apply_time=sushy.APPLY_TIME_ON_RESET)

    @mock.patch.object(redfish_boot.RedfishVirtualMediaBoot, 'prepare_ramdisk',
                       spec_set=True, autospec=True)
    @mock.patch.object(deploy_utils, 'build_agent_options', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    @mock.patch.object(manager_utils, 'node_power_action', autospec=True)
    def test_apply_configuration_no_supported_apply_times(
            self, mock_power_action, mock_get_system, mock_build_agent_options,
            mock_prepare):
        settings = [{'name': 'ProcTurboMode', 'value': 'Disabled'},
                    {'name': 'NicBoot1', 'value': 'NetworkBoot'}]
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            bios = mock_get_system(task.node).bios
            bios.supported_apply_times = None

            task.driver.bios.apply_configuration(task, settings)

            bios.set_attributes.assert_called_once_with(
                {s['name']: s['value'] for s in settings},
                apply_time=None)


@mock.patch('oslo_utils.eventletutils.EventletEvent.wait',
            lambda *args, **kwargs: None)
class RedfishBiosRegistryTestCase(db_base.DbTestCase):

    def setUp(self):
        super(RedfishBiosRegistryTestCase, self).setUp()
        self.config(enabled_bios_interfaces=['redfish'],
                    enabled_hardware_types=['redfish'],
                    enabled_power_interfaces=['redfish'],
                    enabled_boot_interfaces=['redfish-virtual-media'],
                    enabled_management_interfaces=['redfish'])
        self.node = obj_utils.create_test_node(
            self.context, driver='redfish', driver_info=INFO_DICT)

        self.settings = {'SystemModelName': 'UltraSumma',
                         'DcuStreamPrefetcher': 'Enabled',
                         'BootDelay': 10}

        class AttributeField():

            def __init__(self):
                self.name = None
                self.allowable_values = None
                self.attribute_type = None
                self.lower_bound = None
                self.max_length = None
                self.min_length = None
                self.read_only = None
                self.reset_required = None
                self.type = None
                self.unique = None
                self.upper_bound = None

        class AttributeRegistryEntryField():

            def __init__(self, num_entries):
                self.attributes = []
                for _ in range(num_entries):
                    self.attributes.append(AttributeField())

        class AttributeRegistryTest():

            def __init__(self, num_entries):
                self.registry_entries = AttributeRegistryEntryField(
                    num_entries)

        self.registry = AttributeRegistryTest(4)
        self.registry.registry_entries.attributes[0].name = "SystemModelName"
        self.registry.registry_entries.attributes[0].attribute_type = "String"
        self.registry.registry_entries.attributes[0].max_length = 32
        self.registry.registry_entries.attributes[0].read_only = True
        self.registry.registry_entries.attributes[0].unique = True
        self.registry.registry_entries.attributes[1].name =\
            "DcuStreamPrefetcher"
        self.registry.registry_entries.attributes[1].attribute_type =\
            "Enumeration"
        self.registry.registry_entries.attributes[1].read_only = False
        self.registry.registry_entries.attributes[1].allowable_values =\
            [{'ValueName': 'Enabled', 'ValueDisplayName': 'Enabled'},
             {'ValueName': 'Disabled', 'ValueDisplayName': 'Disabled'}]
        self.registry.registry_entries.attributes[2].name = "BootDelay"
        self.registry.registry_entries.attributes[2].attribute_type = "Integer"
        self.registry.registry_entries.attributes[2].lower_bound = 5
        self.registry.registry_entries.attributes[2].upper_bound = 30
        self.registry.registry_entries.attributes[2].reset_required = True
        self.registry.registry_entries.attributes[3].name = "SomeAttribute"
        self.registry.registry_entries.attributes[3].attribute_type = "String"
        self.registry.registry_entries.attributes[3].max_length = 32
        self.registry.registry_entries.attributes[3].read_only = True

        self.expected_no_registry = [
            {'name': 'SystemModelName', 'value': 'UltraSumma'},
            {'name': 'DcuStreamPrefetcher', 'value': 'Enabled'},
            {'name': 'BootDelay', 'value': 10}]

    @mock.patch.object(objects, 'BIOSSettingList', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_cache_bios_registry_save(self, mock_get_system,
                                      mock_setting_list):
        create_list = []
        update_list = []
        delete_list = []
        nochange_list = []
        mock_setting_list.sync_node_setting.return_value = (
            create_list, update_list, delete_list, nochange_list
        )

        expected = [{'name': 'SystemModelName', 'value': 'UltraSumma',
                     'allowable_values': None, 'lower_bound': None,
                     'max_length': 32, 'min_length': None, 'read_only': True,
                     'reset_required': None, 'attribute_type': 'String',
                     'unique': True, 'upper_bound': None},
                    {'name': 'DcuStreamPrefetcher', 'value': 'Enabled',
                     'allowable_values': ['Enabled', 'Disabled'],
                     'lower_bound': None, 'max_length': None,
                     'min_length': None, 'read_only': False,
                     'reset_required': None, 'attribute_type':
                     'Enumeration', 'unique': None, 'upper_bound': None},
                    {'name': 'BootDelay', 'value': 10,
                     'allowable_values': None, 'lower_bound': 5,
                     'max_length': None, 'min_length': None, 'read_only': None,
                     'reset_required': True, 'attribute_type': 'Integer',
                     'unique': None, 'upper_bound': 30}]

        mock_get_system.return_value.bios.attributes = self.settings
        mock_get_system.return_value.bios.get_attribute_registry.\
            return_value = self.registry

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.bios.cache_bios_settings(task)
            mock_get_system.assert_called_once_with(task.node)
            mock_setting_list.sync_node_setting.assert_called_once_with(
                task.context, task.node.id, expected)

    @mock.patch.object(objects, 'BIOSSettingList', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_cache_empty_bios_registry(self, mock_get_system,
                                       mock_setting_list):
        create_list = []
        update_list = []
        delete_list = []
        nochange_list = []
        mock_setting_list.sync_node_setting.return_value = (
            create_list, update_list, delete_list, nochange_list
        )

        mock_get_system.return_value.bios.attributes = self.settings
        mock_get_system.return_value.bios.get_attribute_registry.\
            return_value = {}

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.bios.cache_bios_settings(task)
            mock_setting_list.sync_node_setting.assert_called_once_with(
                task.context, task.node.id, self.expected_no_registry)

    @mock.patch.object(objects, 'BIOSSettingList', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_cache_no_bios_registry(self, mock_get_system,
                                    mock_setting_list):
        create_list = []
        update_list = []
        delete_list = []
        nochange_list = []
        mock_setting_list.sync_node_setting.return_value = (
            create_list, update_list, delete_list, nochange_list
        )

        mock_get_system.return_value.bios.attributes = self.settings
        mock_get_system.return_value.bios.get_attribute_registry.\
            return_value = None

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.bios.cache_bios_settings(task)
            mock_setting_list.sync_node_setting.assert_called_once_with(
                task.context, task.node.id, self.expected_no_registry)

    @mock.patch.object(objects, 'BIOSSettingList', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_cache_exception_bios_registry(self, mock_get_system,
                                           mock_setting_list):
        create_list = []
        update_list = []
        delete_list = []
        nochange_list = []
        mock_setting_list.sync_node_setting.return_value = (
            create_list, update_list, delete_list, nochange_list
        )

        mock_get_system.return_value.bios.attributes = self.settings

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            bios = mock_get_system(task.node).bios
            bios.reset_bios.side_effect = sushy.exceptions.SushyError

            task.driver.bios.cache_bios_settings(task)
            mock_setting_list.sync_node_setting.assert_called_once_with(
                task.context, task.node.id, self.expected_no_registry)



# Copyright 2016 Hewlett Packard Enterprise Development Company LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Test class for Firmware Processor used by iLO management interface."""

import builtins
import io
from unittest import mock
from urllib import parse as urlparse

from oslo_utils import importutils
import proliantutils

from ironic.common import exception
from ironic.drivers.modules.ilo import common as ilo_common
from ironic.drivers.modules.ilo import firmware_processor as ilo_fw_processor
from ironic.tests import base

ilo_error = importutils.try_import('proliantutils.exception')


class FirmwareProcessorTestCase(base.TestCase):

    def setUp(self):
        super(FirmwareProcessorTestCase, self).setUp()
        self.any_url = 'http://netloc/path'
        self.fw_processor_fake = mock.MagicMock(
            parsed_url='set it as required')

    def test_verify_firmware_update_args_throws_for_invalid_update_mode(self):
        # | GIVEN |
        update_firmware_mock = mock.MagicMock()
        firmware_update_args = {'firmware_update_mode': 'invalid_mode',
                                'firmware_images': None}
        # Note(deray): Need to set __name__ attribute explicitly to keep
        # ``functools.wraps`` happy. Passing this to the `name` argument at
        # the time creation of Mock doesn't help.
        update_firmware_mock.__name__ = 'update_firmware_mock'
        wrapped_func = (ilo_fw_processor.
                        verify_firmware_update_args(update_firmware_mock))
        node_fake = mock.MagicMock(uuid='fake_node_uuid')
        task_fake = mock.MagicMock(node=node_fake)
        # | WHEN & THEN |
        self.assertRaises(exception.InvalidParameterValue,
                          wrapped_func,
                          mock.ANY,
                          task_fake,
                          **firmware_update_args)

    def test_verify_firmware_update_args_throws_for_no_firmware_url(self):
        # | GIVEN |
        update_firmware_mock = mock.MagicMock()
        firmware_update_args = {'firmware_update_mode': 'ilo',
                                'firmware_images': []}
        update_firmware_mock.__name__ = 'update_firmware_mock'
        wrapped_func = (ilo_fw_processor.
                        verify_firmware_update_args(update_firmware_mock))
        # | WHEN & THEN |
        self.assertRaises(exception.InvalidParameterValue,
                          wrapped_func,
                          mock.ANY,
                          mock.ANY,
                          **firmware_update_args)

    def test_get_and_validate_firmware_image_info(self):
        # | GIVEN |
        firmware_image_info = {
            'url': self.any_url,
            'checksum': 'b64c8f7799cfbb553d384d34dc43fafe336cc889',
            'component': 'BIOS'
        }
        # | WHEN |
        url, checksum, component = (
            ilo_fw_processor.get_and_validate_firmware_image_info(
                firmware_image_info, 'ilo'))
        # | THEN |
        self.assertEqual(self.any_url, url)
        self.assertEqual('b64c8f7799cfbb553d384d34dc43fafe336cc889', checksum)
        self.assertEqual('bios', component)

    def test_get_and_validate_firmware_image_info_fails_for_missing_parameter(
            self):
        # | GIVEN |
        invalid_firmware_image_info = {
            'url': self.any_url,
            'component': 'bios'
        }
        # | WHEN | & | THEN |
        self.assertRaisesRegex(
            exception.MissingParameterValue, 'checksum',
            ilo_fw_processor.get_and_validate_firmware_image_info,
            invalid_firmware_image_info, 'ilo')

    def test_get_and_validate_firmware_image_info_fails_for_empty_parameter(
            self):
        # | GIVEN |
        invalid_firmware_image_info = {
            'url': self.any_url,
            'checksum': 'valid_checksum',
            'component': ''
        }
        # | WHEN | & | THEN |
        self.assertRaisesRegex(
            exception.MissingParameterValue, 'component',
            ilo_fw_processor.get_and_validate_firmware_image_info,
            invalid_firmware_image_info, 'ilo')

    def test_get_and_validate_firmware_image_info_fails_for_invalid_component(
            self):
        # | GIVEN |
        invalid_firmware_image_info = {
            'url': self.any_url,
            'checksum': 'valid_checksum',
            'component': 'INVALID'
        }
        # | WHEN | & | THEN |
        self.assertRaises(
            exception.InvalidParameterValue,
            ilo_fw_processor.get_and_validate_firmware_image_info,
            invalid_firmware_image_info, 'ilo')

    def test_get_and_validate_firmware_image_info_sum(self):
        # | GIVEN |
        result = None
        firmware_image_info = {
            'url': self.any_url,
            'checksum': 'b64c8f7799cfbb553d384d34dc43fafe336cc889'
        }
        # | WHEN | & | THEN |
        ret_val = ilo_fw_processor.get_and_validate_firmware_image_info(
            firmware_image_info, 'sum')
        self.assertEqual(result, ret_val)

    def test_get_and_validate_firmware_image_info_sum_with_component(self):
        # | GIVEN |
        result = None
        firmware_image_info = {
            'url': self.any_url,
            'checksum': 'b64c8f7799cfbb553d384d34dc43fafe336cc889',
            'components': ['CP02345.exe']
        }
        # | WHEN | & | THEN |
        ret_val = ilo_fw_processor.get_and_validate_firmware_image_info(
            firmware_image_info, 'sum')
        self.assertEqual(result, ret_val)

    def test_get_and_validate_firmware_image_info_sum_invalid_component(
            self):
        # | GIVEN |
        invalid_firmware_image_info = {
            'url': 'any_url',
            'checksum': 'valid_checksum',
            'components': 'INVALID'
        }
        # | WHEN | & | THEN |
        self.assertRaises(
            exception.InvalidParameterValue,
            ilo_fw_processor.get_and_validate_firmware_image_info,
            invalid_firmware_image_info, 'sum')

    def test__validate_sum_components(self):
        result = None
        components = ['CP02345.scexe', 'CP02678.exe']

        ret_val = ilo_fw_processor._validate_sum_components(components)

        self.assertEqual(ret_val, result)

    @mock.patch.object(ilo_fw_processor, 'LOG', autospec=True)
    def test__validate_sum_components_fails(self, LOG_mock):
        components = ['INVALID']

        self.assertRaises(
            exception.InvalidParameterValue,
            ilo_fw_processor._validate_sum_components, components)

        self.assertTrue(LOG_mock.error.called)

    def test_fw_processor_ctor_sets_parsed_url_attrib_of_fw_processor(self):
        # | WHEN |
        fw_processor = ilo_fw_processor.FirmwareProcessor(self.any_url)
        # | THEN |
        self.assertEqual(self.any_url, fw_processor.parsed_url.geturl())

    @mock.patch.object(
        ilo_fw_processor, '_download_file_based_fw_to', autospec=True)
    def test__download_file_based_fw_to_gets_invoked_for_file_based_firmware(
            self, _download_file_based_fw_to_mock):
        # | GIVEN |
        some_file_url = 'file:///some_location/some_firmware_file'
        # | WHEN |
        fw_processor = ilo_fw_processor.FirmwareProcessor(some_file_url)
        fw_processor._download_fw_to('some_target_file')
        # | THEN |
        _download_file_based_fw_to_mock.assert_called_once_with(
            fw_processor, 'some_target_file')

    @mock.patch.object(
        ilo_fw_processor, '_download_http_based_fw_to', autospec=True)
    def test__download_http_based_fw_to_gets_invoked_for_http_based_firmware(
            self, _download_http_based_fw_to_mock):
        # | GIVEN |
        for some_http_url in ('http://netloc/path_to_firmware_file',
                              'https://netloc/path_to_firmware_file'):
            # | WHEN |
            fw_processor = ilo_fw_processor.FirmwareProcessor(some_http_url)
            fw_processor._download_fw_to('some_target_file')
            # | THEN |
            _download_http_based_fw_to_mock.assert_called_once_with(
                fw_processor, 'some_target_file')
            _download_http_based_fw_to_mock.reset_mock()

    @mock.patch.object(
        ilo_fw_processor, '_download_swift_based_fw_to', autospec=True)
    def test__download_swift_based_fw_to_gets_invoked_for_swift_based_firmware(
            self, _download_swift_based_fw_to_mock):
        # | GIVEN |
        some_swift_url = 'swift://containername/objectname'
        # | WHEN |
        fw_processor = ilo_fw_processor.FirmwareProcessor(some_swift_url)
        fw_processor._download_fw_to('some_target_file')
        # | THEN |
        _download_swift_based_fw_to_mock.assert_called_once_with(
            fw_processor, 'some_target_file')

    def test_fw_processor_ctor_throws_exception_with_invalid_firmware_url(
            self):
        # | GIVEN |
        any_invalid_firmware_url = 'any_invalid_url'
        # | WHEN | & | THEN |
        self.assertRaises(exception.InvalidParameterValue,
                          ilo_fw_processor.FirmwareProcessor,
                          any_invalid_firmware_url)

    @mock.patch.object(ilo_fw_processor, 'tempfile', autospec=True)
    @mock.patch.object(ilo_fw_processor, 'os', autospec=True)
    @mock.patch.object(ilo_fw_processor, 'shutil', autospec=True)
    @mock.patch.object(ilo_common, 'verify_image_checksum',
                       spec_set=True, autospec=True)
    @mock.patch.object(
        ilo_fw_processor, '_extract_fw_from_file', autospec=True)
    def test_process_fw_on_calls__download_fw_to(
            self, _extract_fw_from_file_mock, verify_checksum_mock,
            shutil_mock, os_mock, tempfile_mock):
        # | GIVEN |
        fw_processor = ilo_fw_processor.FirmwareProcessor(self.any_url)
        # Now mock the __download_fw_to method of fw_processor instance
        _download_fw_to_mock = mock.MagicMock()
        fw_processor._download_fw_to = _download_fw_to_mock

        expected_return_location = (ilo_fw_processor.FirmwareImageLocation(
            'some_location/file', 'file'))
        _extract_fw_from_file_mock.return_value = (expected_return_location,
                                                   True)
        node_mock = mock.ANY
        checksum_fake = mock.ANY
        # | WHEN |
        actual_return_location = fw_processor.process_fw_on(node_mock,
                                                            checksum_fake)
        # | THEN |
        _download_fw_to_mock.assert_called_once_with(
            os_mock.path.join.return_value)
        self.assertEqual(expected_return_location.fw_image_location,
                         actual_return_location.fw_image_location)
        self.assertEqual(expected_return_location.fw_image_filename,
                         actual_return_location.fw_image_filename)

    @mock.patch.object(ilo_fw_processor, 'tempfile', autospec=True)
    @mock.patch.object(ilo_fw_processor, 'os', autospec=True)
    @mock.patch.object(ilo_fw_processor, 'shutil', autospec=True)
    @mock.patch.object(ilo_common, 'verify_image_checksum',
                       spec_set=True, autospec=True)
    @mock.patch.object(
        ilo_fw_processor, '_extract_fw_from_file', autospec=True)
    def test_process_fw_on_verifies_checksum_of_downloaded_fw_file(
            self, _extract_fw_from_file_mock, verify_checksum_mock,
            shutil_mock, os_mock, tempfile_mock):
        # | GIVEN |
        fw_processor = ilo_fw_processor.FirmwareProcessor(self.any_url)
        # Now mock the __download_fw_to method of fw_processor instance
        _download_fw_to_mock = mock.MagicMock()
        fw_processor._download_fw_to = _download_fw_to_mock

        expected_return_location = (ilo_fw_processor.FirmwareImageLocation(
            'some_location/file', 'file'))
        _extract_fw_from_file_mock.return_value = (expected_return_location,
                                                   True)
        node_mock = mock.ANY
        checksum_fake = mock.ANY
        # | WHEN |
        actual_return_location = fw_processor.process_fw_on(node_mock,
                                                            checksum_fake)
        # | THEN |
        _download_fw_to_mock.assert_called_once_with(
            os_mock.path.join.return_value)
        verify_checksum_mock.assert_called_once_with(
            os_mock.path.join.return_value, checksum_fake)
        self.assertEqual(expected_return_location.fw_image_location,
                         actual_return_location.fw_image_location)
        self.assertEqual(expected_return_location.fw_image_filename,
                         actual_return_location.fw_image_filename)

    @mock.patch.object(ilo_fw_processor, 'tempfile', autospec=True)
    @mock.patch.object(ilo_fw_processor, 'os', autospec=True)
    @mock.patch.object(ilo_fw_processor, 'shutil', autospec=True)
    @mock.patch.object(ilo_common, 'verify_image_checksum',
                       spec_set=True, autospec=True)
    def test_process_fw_on_throws_error_if_checksum_validation_fails(
            self, verify_checksum_mock, shutil_mock, os_mock, tempfile_mock):
        # | GIVEN |
        fw_processor = ilo_fw_processor.FirmwareProcessor(self.any_url)
        # Now mock the __download_fw_to method of fw_processor instance
        _download_fw_to_mock = mock.MagicMock()
        fw_processor._download_fw_to = _download_fw_to_mock

        verify_checksum_mock.side_effect = exception.ImageRefValidationFailed(
            image_href='some image',
            reason='checksum verification failed')
        node_mock = mock.ANY
        checksum_fake = mock.ANY
        # | WHEN | & | THEN |
        self.assertRaises(exception.ImageRefValidationFailed,
                          fw_processor.process_fw_on,
                          node_mock,
                          checksum_fake)
        shutil_mock.rmtree.assert_called_once_with(
            tempfile_mock.mkdtemp(), ignore_errors=True)

    @mock.patch.object(ilo_fw_processor, 'tempfile', autospec=True)
    @mock.patch.object(ilo_fw_processor, 'os', autospec=True)
    @mock.patch.object(ilo_fw_processor, 'shutil', autospec=True)
    @mock.patch.object(ilo_common, 'verify_image_checksum',
                       spec_set=True, autospec=True)
    @mock.patch.object(
        ilo_fw_processor, '_extract_fw_from_file', autospec=True)
    def test_process_fw_on_calls__extract_fw_from_file(
            self, _extract_fw_from_file_mock, verify_checksum_mock,
            shutil_mock, os_mock, tempfile_mock):
        # | GIVEN |
        fw_processor = ilo_fw_processor.FirmwareProcessor(self.any_url)
        # Now mock the __download_fw_to method of fw_processor instance
        _download_fw_to_mock = mock.MagicMock()
        fw_processor._download_fw_to = _download_fw_to_mock

        expected_return_location = (ilo_fw_processor.FirmwareImageLocation(
            'some_location/file', 'file'))
        _extract_fw_from_file_mock.return_value = (expected_return_location,
                                                   True)
        node_mock = mock.ANY
        checksum_fake = mock.ANY
        # | WHEN |
        actual_return_location = fw_processor.process_fw_on(node_mock,
                                                            checksum_fake)
        # | THEN |
        _extract_fw_from_file_mock.assert_called_once_with(
            node_mock, os_mock.path.join.return_value)
        self.assertEqual(expected_return_location.fw_image_location,
                         actual_return_location.fw_image_location)
        self.assertEqual(expected_return_location.fw_image_filename,
                         actual_return_location.fw_image_filename)
        shutil_mock.rmtree.assert_called_once_with(
            tempfile_mock.mkdtemp(), ignore_errors=True)

    @mock.patch.object(builtins, 'open', autospec=True)
    @mock.patch.object(
        ilo_fw_processor.image_service, 'FileImageService', autospec=True)
    def test__download_file_based_fw_to_copies_file_to_target(
            self, file_image_service_mock, open_mock):
        # | GIVEN |
        fd_mock = mock.MagicMock(spec=io.BytesIO)
        open_mock.return_value = fd_mock
        fd_mock.__enter__.return_value = fd_mock
        any_file_based_firmware_file = 'file:///tmp/any_file_path'
        firmware_file_path = '/tmp/any_file_path'
        self.fw_processor_fake.parsed_url = urlparse.urlparse(
            any_file_based_firmware_file)
        # | WHEN |
        ilo_fw_processor._download_file_based_fw_to(self.fw_processor_fake,
                                                    'target_file')
        # | THEN |
        file_image_service_mock.return_value.download.assert_called_once_with(
            firmware_file_path, fd_mock)

    @mock.patch.object(builtins, 'open', autospec=True)
    @mock.patch.object(ilo_fw_processor, 'image_service', autospec=True)
    def test__download_http_based_fw_to_downloads_the_fw_file(
            self, image_service_mock, open_mock):
        # | GIVEN |
        fd_mock = mock.MagicMock(spec=io.BytesIO)
        open_mock.return_value = fd_mock
        fd_mock.__enter__.return_value = fd_mock
        any_http_based_firmware_file = 'http://netloc/path_to_firmware_file'
        any_target_file = 'any_target_file'
        self.fw_processor_fake.parsed_url = urlparse.urlparse(
            any_http_based_firmware_file)
        # | WHEN |
        ilo_fw_processor._download_http_based_fw_to(self.fw_processor_fake,
                                                    any_target_file)
        # | THEN |
        image_service_mock.HttpImageService().download.assert_called_once_with(
            any_http_based_firmware_file, fd_mock)

    @mock.patch.object(ilo_fw_processor, 'urlparse', autospec=True)
    @mock.patch.object(
        ilo_fw_processor, '_download_http_based_fw_to', autospec=True)
    @mock.patch.object(ilo_fw_processor, 'swift', autospec=True)
    def test__download_swift_based_fw_to_creates_temp_url(
            self, swift_mock, _download_http_based_fw_to_mock, urlparse_mock):
        # | GIVEN |
        swift_based_firmware_files = [
            'swift://containername/objectname',
            'swift://containername/pseudo-folder/objectname'
        ]
        for swift_firmware_file in swift_based_firmware_files:
            # | WHEN |
            self.fw_processor_fake.parsed_url = (urlparse.
                                                 urlparse(swift_firmware_file))
            ilo_fw_processor._download_swift_based_fw_to(
                self.fw_processor_fake, 'any_target_file')
        # | THEN |
        expected_temp_url_call_args_list = [
            mock.call('containername', 'objectname', mock.ANY),
            mock.call('containername', 'pseudo-folder/objectname', mock.ANY)
        ]
        actual_temp_url_call_args_list = (
            swift_mock.SwiftAPI().get_temp_url.call_args_list)
        self.assertEqual(expected_temp_url_call_args_list,
                         actual_temp_url_call_args_list)

    @mock.patch.object(urlparse, 'urlparse', autospec=True)
    @mock.patch.object(
        ilo_fw_processor, '_download_http_based_fw_to', autospec=True)
    @mock.patch.object(ilo_fw_processor, 'swift', autospec=True)
    def test__download_swift_based_fw_to_calls__download_http_based_fw_to(
            self, swift_mock, _download_http_based_fw_to_mock, urlparse_mock):
        """_download_swift_based_fw_to invokes _download_http_based_fw_to

        _download_swift_based_fw_to makes a call to _download_http_based_fw_to
        in turn with temp url set as the url attribute of fw_processor instance
        """
        # | GIVEN |
        any_swift_based_firmware_file = 'swift://containername/objectname'
        any_target_file = 'any_target_file'
        self.fw_processor_fake.parsed_url = urlparse.urlparse(
            any_swift_based_firmware_file)
        urlparse_mock.reset_mock()
        # | WHEN |
        ilo_fw_processor._download_swift_based_fw_to(self.fw_processor_fake,
                                                     any_target_file)
        # | THEN |
        _download_http_based_fw_to_mock.assert_called_once_with(
            self.fw_processor_fake, any_target_file)
        urlparse_mock.assert_called_once_with(
            swift_mock.SwiftAPI().get_temp_url.return_value)
        self.assertEqual(
            urlparse_mock.return_value, self.fw_processor_fake.parsed_url)

    @mock.patch.object(ilo_fw_processor, 'ilo_common', autospec=True)
    def test__extract_fw_from_file_calls_process_firmware_image(
            self, ilo_common_mock):
        # | GIVEN |
        if not mock._is_instance_mock(proliantutils.utils):
            mock.patch.object(proliantutils.utils, 'process_firmware_image',
                              autospec=True).start()
        utils_mock = proliantutils.utils.process_firmware_image
        node_mock = mock.MagicMock(uuid='fake_node_uuid')
        any_target_file = 'any_target_file'
        ilo_object_mock = ilo_common_mock.get_ilo_object.return_value
        utils_mock.return_value = ('some_location', True, True)
        # | WHEN |
        ilo_fw_processor._extract_fw_from_file(node_mock, any_target_file)
        # | THEN |
        utils_mock.assert_called_once_with(any_target_file, ilo_object_mock)

    @mock.patch.object(ilo_fw_processor, 'ilo_common', autospec=True)
    def test__extract_fw_from_file_doesnt_upload_firmware(
            self, ilo_common_mock):
        # | GIVEN |
        if not mock._is_instance_mock(proliantutils.utils):
            mock.patch.object(proliantutils.utils, 'process_firmware_image',
                              autospec=True).start()
        utils_mock = proliantutils.utils.process_firmware_image
        node_mock = mock.MagicMock(uuid='fake_node_uuid')
        any_target_file = 'any_target_file'
        utils_mock.return_value = (
            'some_location/some_fw_file', False, True)
        # | WHEN |
        ilo_fw_processor._extract_fw_from_file(node_mock, any_target_file)
        # | THEN |
        ilo_common_mock.copy_image_to_web_server.assert_not_called()

    @mock.patch.object(ilo_fw_processor, 'ilo_common', autospec=True)
    @mock.patch.object(ilo_fw_processor, '_remove_file_based_me',
                       autospec=True)
    def test__extract_fw_from_file_sets_loc_obj_remove_to_file_if_no_upload(
            self, _remove_mock, ilo_common_mock):
        # | GIVEN |
        if not mock._is_instance_mock(proliantutils.utils):
            mock.patch.object(proliantutils.utils, 'process_firmware_image',
                              autospec=True).start()
        utils_mock = proliantutils.utils.process_firmware_image
        node_mock = mock.MagicMock(uuid='fake_node_uuid')
        any_target_file = 'any_target_file'
        utils_mock.return_value = (
            'some_location/some_fw_file', False, True)
        # | WHEN |
        location_obj, is_different_file = (
            ilo_fw_processor._extract_fw_from_file(node_mock, any_target_file))
        location_obj.remove()
        # | THEN |
        _remove_mock.assert_called_once_with(location_obj)

    @mock.patch.object(ilo_fw_processor, 'ilo_common', autospec=True)
    def test__extract_fw_from_file_uploads_firmware_to_webserver(
            self, ilo_common_mock):
        # | GIVEN |
        if not mock._is_instance_mock(proliantutils.utils):
            mock.patch.object(proliantutils.utils, 'process_firmware_image',
                              autospec=True).start()
        utils_mock = proliantutils.utils.process_firmware_image
        node_mock = mock.MagicMock(uuid='fake_node_uuid')
        any_target_file = 'any_target_file'
        utils_mock.return_value = (
            'some_location/some_fw_file', True, True)
        self.config(use_web_server_for_images=True, group='ilo')
        # | WHEN |
        ilo_fw_processor._extract_fw_from_file(node_mock, any_target_file)
        # | THEN |
        ilo_common_mock.copy_image_to_web_server.assert_called_once_with(
            'some_location/some_fw_file', 'some_fw_file')

    @mock.patch.object(ilo_fw_processor, 'ilo_common', autospec=True)
    @mock.patch.object(ilo_fw_processor, '_remove_webserver_based_me',
                       autospec=True)
    def test__extract_fw_from_file_sets_loc_obj_remove_to_webserver(
            self, _remove_mock, ilo_common_mock):
        # | GIVEN |
        if not mock._is_instance_mock(proliantutils.utils):
            mock.patch.object(proliantutils.utils, 'process_firmware_image',
                              autospec=True).start()
        utils_mock = proliantutils.utils.process_firmware_image
        node_mock = mock.MagicMock(uuid='fake_node_uuid')
        any_target_file = 'any_target_file'
        utils_mock.return_value = (
            'some_location/some_fw_file', True, True)
        self.config(use_web_server_for_images=True, group='ilo')
        # | WHEN |
        location_obj, is_different_file = (
            ilo_fw_processor._extract_fw_from_file(node_mock, any_target_file))
        location_obj.remove()
        # | THEN |
        _remove_mock.assert_called_once_with(location_obj)

    @mock.patch.object(ilo_fw_processor, 'ilo_common', autospec=True)
    def test__extract_fw_from_file_uploads_firmware_to_swift(
            self, ilo_common_mock):
        # | GIVEN |
        if not mock._is_instance_mock(proliantutils.utils):
            mock.patch.object(proliantutils.utils, 'process_firmware_image',
                              autospec=True).start()
        utils_mock = proliantutils.utils.process_firmware_image
        node_mock = mock.MagicMock(uuid='fake_node_uuid')
        any_target_file = 'any_target_file'
        utils_mock.return_value = (
            'some_location/some_fw_file', True, True)
        self.config(use_web_server_for_images=False, group='ilo')
        # | WHEN |
        ilo_fw_processor._extract_fw_from_file(node_mock, any_target_file)
        # | THEN |
        ilo_common_mock.copy_image_to_swift.assert_called_once_with(
            'some_location/some_fw_file', 'some_fw_file')

    @mock.patch.object(ilo_fw_processor, 'ilo_common', autospec=True)
    @mock.patch.object(ilo_fw_processor, '_remove_swift_based_me',
                       autospec=True)
    def test__extract_fw_from_file_sets_loc_obj_remove_to_swift(
            self, _remove_mock, ilo_common_mock):
        # | GIVEN |
        if not mock._is_instance_mock(proliantutils.utils):
            mock.patch.object(proliantutils.utils, 'process_firmware_image',
                              autospec=True).start()
        utils_mock = proliantutils.utils.process_firmware_image
        node_mock = mock.MagicMock(uuid='fake_node_uuid')
        any_target_file = 'any_target_file'
        utils_mock.return_value = (
            'some_location/some_fw_file', True, True)
        self.config(use_web_server_for_images=False, group='ilo')
        # | WHEN |
        location_obj, is_different_file = (
            ilo_fw_processor._extract_fw_from_file(node_mock, any_target_file))
        location_obj.remove()
        # | THEN |
        _remove_mock.assert_called_once_with(location_obj)

    def test_fw_img_loc_sets_these_attributes(self):
        # | GIVEN |
        any_loc = 'some_location/some_fw_file'
        any_s_filename = 'some_fw_file'
        # | WHEN |
        location_obj = ilo_fw_processor.FirmwareImageLocation(
            any_loc, any_s_filename)
        # | THEN |
        self.assertEqual(any_loc, location_obj.fw_image_location)
        self.assertEqual(any_s_filename, location_obj.fw_image_filename)

    @mock.patch.object(ilo_fw_processor, 'ilo_common', autospec=True)
    def test__remove_file_based_me(
            self, ilo_common_mock):
        # | GIVEN |
        fw_img_location_obj_fake = mock.MagicMock()
        # | WHEN |
        ilo_fw_processor._remove_file_based_me(fw_img_location_obj_fake)
        # | THEN |
        (ilo_common_mock.remove_single_or_list_of_files.
         assert_called_with(fw_img_location_obj_fake.fw_image_location))

    @mock.patch.object(ilo_fw_processor, 'ilo_common', autospec=True)
    def test__remove_swift_based_me(self, ilo_common_mock):
        # | GIVEN |
        fw_img_location_obj_fake = mock.MagicMock()
        # | WHEN |
        ilo_fw_processor._remove_swift_based_me(fw_img_location_obj_fake)
        # | THEN |
        (ilo_common_mock.remove_image_from_swift.assert_called_with(
            fw_img_location_obj_fake.fw_image_filename, "firmware update"))

    @mock.patch.object(ilo_fw_processor, 'ilo_common', autospec=True)
    def test__remove_webserver_based_me(self, ilo_common_mock):
        # | GIVEN |
        fw_img_location_obj_fake = mock.MagicMock()
        # | WHEN |
        ilo_fw_processor._remove_webserver_based_me(fw_img_location_obj_fake)
        # | THEN |
        (ilo_common_mock.remove_image_from_web_server.assert_called_with(
            fw_img_location_obj_fake.fw_image_filename))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ironic.conductor import task_manager
from ironic.drivers.modules import agent
from ironic.drivers.modules import ipmitool
from ironic.drivers.modules import noop
from ironic.drivers.modules import noop_mgmt
from ironic.drivers.modules import pxe
from ironic.drivers.modules.storage import cinder
from ironic.drivers.modules.storage import noop as noop_storage
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils


class IPMIHardwareTestCase(db_base.DbTestCase):

    def setUp(self):
        super(IPMIHardwareTestCase, self).setUp()
        self.config(enabled_hardware_types=['ipmi'],
                    enabled_power_interfaces=['ipmitool'],
                    enabled_management_interfaces=['ipmitool', 'noop'],
                    enabled_raid_interfaces=['no-raid', 'agent'],
                    enabled_console_interfaces=['no-console'],
                    enabled_vendor_interfaces=['ipmitool', 'no-vendor'])

    def _validate_interfaces(self, task, **kwargs):
        self.assertIsInstance(
            task.driver.management,
            kwargs.get('management', ipmitool.IPMIManagement))
        self.assertIsInstance(
            task.driver.power,
            kwargs.get('power', ipmitool.IPMIPower))
        self.assertIsInstance(
            task.driver.boot,
            kwargs.get('boot', pxe.PXEBoot))
        self.assertIsInstance(
            task.driver.deploy,
            kwargs.get('deploy', agent.AgentDeploy))
        self.assertIsInstance(
            task.driver.console,
            kwargs.get('console', noop.NoConsole))
        self.assertIsInstance(
            task.driver.raid,
            kwargs.get('raid', noop.NoRAID))
        self.assertIsInstance(
            task.driver.vendor,
            kwargs.get('vendor', ipmitool.VendorPassthru))
        self.assertIsInstance(
            task.driver.storage,
            kwargs.get('storage', noop_storage.NoopStorage))
        self.assertIsInstance(
            task.driver.rescue,
            kwargs.get('rescue', noop.NoRescue))

    def test_default_interfaces(self):
        node = obj_utils.create_test_node(self.context, driver='ipmi')
        with task_manager.acquire(self.context, node.id) as task:
            self._validate_interfaces(task)

    def test_override_with_shellinabox(self):
        self.config(enabled_console_interfaces=['ipmitool-shellinabox',
                                                'ipmitool-socat'])
        node = obj_utils.create_test_node(
            self.context, driver='ipmi',
            raid_interface='agent',
            console_interface='ipmitool-shellinabox',
            vendor_interface='no-vendor')
        with task_manager.acquire(self.context, node.id) as task:
            self._validate_interfaces(
                task,
                console=ipmitool.IPMIShellinaboxConsole,
                raid=agent.AgentRAID,
                vendor=noop.NoVendor)

    def test_override_with_cinder_storage(self):
        self.config(enabled_storage_interfaces=['noop', 'cinder'])
        node = obj_utils.create_test_node(
            self.context, driver='ipmi',
            storage_interface='cinder')
        with task_manager.acquire(self.context, node.id) as task:
            self._validate_interfaces(task, storage=cinder.CinderStorage)

    def test_override_with_agent_rescue(self):
        self.config(enabled_rescue_interfaces=['no-rescue', 'agent'])
        node = obj_utils.create_test_node(
            self.context, driver='ipmi',
            rescue_interface='agent')
        with task_manager.acquire(self.context, node.id) as task:
            self._validate_interfaces(task, rescue=agent.AgentRescue)

    def test_override_with_noop_mgmt(self):
        self.config(enabled_management_interfaces=['ipmitool', 'noop'])
        node = obj_utils.create_test_node(
            self.context, driver='ipmi',
            management_interface='noop')
        with task_manager.acquire(self.context, node.id) as task:
            self._validate_interfaces(task,
                                      management=noop_mgmt.NoopManagement)



# Copyright 2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from http import client as http_client
import os
import ssl

from ironic_lib import metrics_utils
from oslo_log import log
from oslo_serialization import jsonutils
from oslo_utils import strutils
import requests
import tenacity

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import utils
from ironic.conf import CONF

LOG = log.getLogger(__name__)

METRICS = metrics_utils.get_metrics_logger(__name__)

DEFAULT_IPA_PORTAL_PORT = 3260

REBOOT_COMMAND = 'run_image'


def get_client(task):
    """Get client for this node."""
    try:
        return task.cached_agent_client
    except AttributeError:
        task.cached_agent_client = AgentClient()
        return task.cached_agent_client


def get_command_error(command):
    """Extract an error string from the command result.

    :param command: Command information from the agent.
    :return: Error string.
    """
    error = command.get('command_error')
    if error is None:
        LOG.error('Agent returned invalid response: missing command_error in '
                  '%s', command)
        return _('Invalid agent response')

    if isinstance(error, dict):
        return error.get('details') or error.get('message') or str(error)
    else:
        return error


def _sanitize_for_logging(var):
    if not var:
        return var
    elif isinstance(var, str):
        return strutils.mask_password(var)
    else:
        return utils.remove_large_keys(strutils.mask_dict_password(var))


class AgentClient(object):
    """Client for interacting with nodes via a REST API."""
    @METRICS.timer('AgentClient.__init__')
    def __init__(self):
        self.session = requests.Session()
        self.session.headers.update({'Content-Type': 'application/json'})

    def _get_command_url(self, node):
        """Get URL endpoint for agent command request"""
        agent_url = node.driver_internal_info.get('agent_url')
        if not agent_url:
            raise exception.AgentConnectionFailed(_('Agent driver requires '
                                                    'agent_url in '
                                                    'driver_internal_info'))
        return ('%(agent_url)s/%(api_version)s/commands/' %
                {'agent_url': agent_url,
                 'api_version': CONF.agent.agent_api_version})

    def _get_command_body(self, method, params):
        """Generate command body from method and params"""
        return jsonutils.dumps({
            'name': method,
            'params': params,
        })

    def _get_verify(self, node):
        # False is a valid value, don't use 'or'
        value = node.driver_internal_info.get('agent_verify_ca')
        if value is None:
            value = node.driver_info.get('agent_verify_ca')
        if value is None:
            value = CONF.agent.verify_ca

        if isinstance(value, str):
            try:
                value = strutils.bool_from_string(value, strict=True)
            except ValueError:
                if not os.path.exists(value):
                    raise exception.InvalidParameterValue(
                        _('Agent CA %s is neither a path nor a boolean')
                        % value)
        return value

    def _raise_if_typeerror(self, result, node, method):
        error = result.get('command_error')
        if error and error.get('type') == 'TypeError':
            LOG.error('Agent command %(method)s for node %(node)s failed. '
                      'Internal TypeError detected: Error %(error)s',
                      {'method': method, 'node': node.uuid, 'error': error})
            raise exception.AgentAPIError(node=node.uuid,
                                          status=error.get('code'),
                                          error=get_command_error(result))

    @METRICS.timer('AgentClient._wait_for_command')
    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(
            exception.AgentCommandTimeout),
        stop=tenacity.stop_after_attempt(CONF.agent.command_wait_attempts),
        wait=tenacity.wait_fixed(CONF.agent.command_wait_interval),
        reraise=True)
    def _wait_for_command(self, node, method):
        """Wait for a command to complete.

        :param node: A Node object.
        :param method: A string represents the command executed by agent.
        :raises: AgentCommandTimeout if timeout is reached.
        """
        # NOTE(dtantsur): this function uses AgentCommandTimeout on every
        # failure, but unless the timeout is reached, the exception is caught
        # and retried by the @retry decorator above.
        result = self.get_last_command_status(node, method)
        if result is None:
            raise exception.AgentCommandTimeout(command=method, node=node.uuid)

        if result.get('command_status') == 'RUNNING':
            LOG.debug('Command %(cmd)s has not finished yet for node %(node)s',
                      {'cmd': method, 'node': node.uuid})
            raise exception.AgentCommandTimeout(command=method, node=node.uuid)
        else:
            LOG.debug('Command %(cmd)s has finished for node %(node)s with '
                      'result %(result)s',
                      {'cmd': method, 'node': node.uuid, 'result': result})
            self._raise_if_typeerror(result, node, method)
            return result

    @METRICS.timer('AgentClient._command')
    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(
            exception.AgentConnectionFailed),
        stop=tenacity.stop_after_attempt(CONF.agent.max_command_attempts),
        reraise=True)
    def _command(self, node, method, params, wait=False, poll=False):
        """Sends command to agent.

        :param node: A Node object.
        :param method: A string represents the command to be executed by
                       agent.
        :param params: A dictionary containing params used to form the request
                       body.
        :param wait: True to wait for the command to finish executing, False
                     otherwise.
        :param poll: Whether to poll the command until completion. Provides
                     a better alternative to `wait` for long-running commands.
        :raises: IronicException when failed to issue the request or there was
                 a malformed response from the agent.
        :raises: AgentAPIError when agent failed to execute specified command.
        :raises: AgentInProgress when the command fails to execute as the agent
                 is presently executing the prior command.
        :returns: A dict containing command result from agent, see
                  get_commands_status for a sample.
        """
        assert not (wait and poll)

        url = self._get_command_url(node)
        body = self._get_command_body(method, params)
        request_params = {
            'wait': str(wait).lower()
        }
        agent_token = node.driver_internal_info.get('agent_secret_token')
        if agent_token:
            request_params['agent_token'] = agent_token
        LOG.debug('Executing agent command %(method)s for node %(node)s '
                  'with params %(params)s',
                  {'node': node.uuid, 'method': method,
                   'params': _sanitize_for_logging(request_params)})

        try:
            response = self.session.post(
                url, params=request_params, data=body,
                verify=self._get_verify(node),
                timeout=CONF.agent.command_timeout)
        except (requests.ConnectionError, requests.Timeout, ssl.SSLError) as e:
            result = self._handle_timeout_on_command_execution(node, method,
                                                               params, e)
            response = None
        except requests.RequestException as e:
            msg = (_('Error invoking agent command %(method)s for node '
                     '%(node)s. Error: %(error)s') %
                   {'method': method, 'node': node.uuid, 'error': e})
            LOG.error(msg)
            raise exception.IronicException(msg)

        if response is not None:
            # TODO(russellhaering): real error handling
            try:
                result = response.json()
            except ValueError:
                msg = _(
                    'Unable to decode response as JSON.\n'
                    'Request URL: %(url)s\nRequest body: "%(body)s"\n'
                    'Response status code: %(code)s\n'
                    'Response: "%(response)s"'
                ) % ({'response': response.text, 'body': body, 'url': url,
                      'code': response.status_code})
                LOG.error(msg)
                raise exception.IronicException(msg)

        error = result.get('command_error')
        LOG.debug('Agent command %(method)s for node %(node)s returned '
                  'result %(res)s, error %(error)s, HTTP status code %(code)s',
                  {'node': node.uuid, 'method': method,
                   'res': _sanitize_for_logging(result.get('command_result')),
                   'error': error,
                   'code': response.status_code if response is not None
                   else 'unknown'})
        if (response is not None
                and response.status_code >= http_client.BAD_REQUEST):
            faultstring = result.get('faultstring')
            if 'agent_token' in faultstring:
                LOG.error('Agent command %(method)s for node %(node)s '
                          'failed. Expected 2xx HTTP status code, got '
                          '%(code)d. Error suggests an older ramdisk '
                          'which does not support ``agent_token``. '
                          'This is a fatal error.',
                          {'method': method, 'node': node.uuid,
                           'code': response.status_code})
            else:
                LOG.error('Agent command %(method)s for node %(node)s failed. '
                          'Expected 2xx HTTP status code, got %(code)d.',
                          {'method': method, 'node': node.uuid,
                           'code': response.status_code})
            if (response.status_code == http_client.CONFLICT
                or 'agent is busy' in faultstring.lower()):
                # HTTP 409 check as an explicit check of if the agent
                # is already busy.
                # NOTE(TheJulia): The agent sends upper case A as of
                # late victoria, but lower case the entire message
                # for compatability with pre-late victoria agents
                # which returns HTTP 409.
                raise exception.AgentInProgress(node=node.uuid,
                                                command=method,
                                                error=faultstring)
            raise exception.AgentAPIError(node=node.uuid,
                                          status=response.status_code,
                                          error=faultstring)

        self._raise_if_typeerror(result, node, method)

        if poll:
            result = self._wait_for_command(node, method)

        return result

    @METRICS.timer('AgentClient.get_commands_status')
    def get_commands_status(self, node, retry_connection=True,
                            expect_errors=False):
        """Get command status from agent.

        :param node: A Node object.
        :param retry_connection: Whether to retry connection problems.
        :param expect_errors: If True, do not log connection problems as
            errors.
        :return: A list of command results, each result is related to a
            command been issued to agent. A typical result can be:

            ::

              {
                'command_name': <command name related to the result>,
                'command_params': <params related with the command>,
                'command_status': <current command status,
                                  e.g. 'RUNNING', 'SUCCEEDED', 'FAILED'>,
                'command_error': <error message if command execution
                                 failed>,
                'command_result': <command result if command execution
                                  succeeded, the value is command specific,
                                  e.g.:
                                  * a dictionary containing keys clean_result
                                    and clean_step for the command
                                    clean.execute_clean_step;
                                  * a dictionary containing keys deploy_result
                                    and deploy_step for the command
                                    deploy.execute_deploy_step;
                                  * a string representing result message for
                                    the command standby.cache_image;
                                  * None for the command standby.sync.>
              }
        """
        url = self._get_command_url(node)
        LOG.debug('Fetching status of agent commands for node %s', node.uuid)

        def _get():
            try:
                return self.session.get(url,
                                        verify=self._get_verify(node),
                                        timeout=CONF.agent.command_timeout)
            except (requests.ConnectionError, requests.Timeout) as e:
                msg = (_('Failed to connect to the agent running on node '
                         '%(node)s to collect commands status. '
                         'Error: %(error)s') %
                       {'node': node.uuid, 'error': e})
                logging_call = LOG.debug if expect_errors else LOG.error
                logging_call(msg)
                raise exception.AgentConnectionFailed(reason=msg)

        if retry_connection:
            _get = tenacity.retry(
                retry=tenacity.retry_if_exception_type(
                    exception.AgentConnectionFailed),
                stop=tenacity.stop_after_attempt(
                    CONF.agent.max_command_attempts),
                reraise=True)(_get)

        result = _get().json()['commands']
        status = '; '.join('%(cmd)s: result "%(res)s", error "%(err)s"' %
                           {'cmd': r.get('command_name'),
                            'res': _sanitize_for_logging(
                                r.get('command_result')),
                            'err': r.get('command_error')}
                           for r in result)
        LOG.debug('Status of agent commands for node %(node)s: %(status)s',
                  {'node': node.uuid, 'status': status})
        return result

    def _status_if_last_command_matches(self, node, method, params):
        """Return the status of the given command if it's the last running."""
        try:
            method = method.split('.', 1)[1]
        except IndexError:
            pass

        commands = self.get_commands_status(node)
        if not commands:
            return None

        # TODO(dtantsur): a more reliable way to detect repeated execution
        # would be to pass a sort of require ID to the agent.

        command = commands[-1]
        if command['command_name'] != method:
            LOG.debug('Command %(cmd)s is not currently executing, the last '
                      'command is %(curr)s',
                      {'cmd': method, 'curr': command['command_name']})
            return None

        if command['command_status'] != 'RUNNING':
            LOG.debug('Command %(cmd)s is not currently executing, its status '
                      'is %(curr)s',
                      {'cmd': method, 'curr': command['command_status']})
            return None

        return command

    def _handle_timeout_on_command_execution(self, node, method, params,
                                             error):
        result = None
        # NOTE(dtantsur): it is possible, especially with eventlet+TLS, that
        # agent receives a command but fails to return the result to Ironic.
        # To avoid a failure, check if the last command is the one we're trying
        # to execute.
        try:
            result = self._status_if_last_command_matches(node, method, params)
        except Exception as e:
            msg = (_('Failed to connect to the agent running on node '
                     '%(node)s for checking the last command status '
                     'after failing to invoke command %(method)s. '
                     'Error: %(error)s') %
                   {'node': node.uuid, 'method': method, 'error': e})
            LOG.error(msg)

        if result is None:
            msg = (_('Failed to connect to the agent running on node %(node)s '
                     'for invoking command %(method)s. Error: %(error)s') %
                   {'node': node.uuid, 'method': method, 'error': error})
            LOG.error(msg)
            raise exception.AgentConnectionFailed(reason=msg)

        return result

    def get_last_command_status(self, node, method):
        """Get the last status for the given command.

        :param node: A Node object.
        :param method: Command name.
        :returns: A dict containing command status from agent or None
            if the command was not found.
        """
        try:
            method = method.split('.', 1)[1]
        except IndexError:
            pass

        commands = self.get_commands_status(node)
        try:
            return next(c for c in reversed(commands)
                        if c.get('command_name') == method)
        except StopIteration:
            LOG.debug('Command %(cmd)s is not in the executing commands list '
                      'for node %(node)s',
                      {'cmd': method, 'node': node.uuid})
            return None

    @METRICS.timer('AgentClient.install_bootloader')
    def install_bootloader(self, node, root_uuid, target_boot_mode,
                           efi_system_part_uuid=None,
                           prep_boot_part_uuid=None,
                           software_raid=False):
        """Install a boot loader on the image.

        :param node: A node object.
        :param root_uuid: The UUID of the root partition.
        :param target_boot_mode: The target deployment boot mode.
        :param efi_system_part_uuid: The UUID of the efi system partition
               where the bootloader will be installed to, only used for uefi
               boot mode.
        :param prep_boot_part_uuid: The UUID of the PReP Boot partition where
               the bootloader will be installed to when local booting a
               partition image on a ppc64* system.
        :raises: IronicException when failed to issue the request or there was
                 a malformed response from the agent.
        :raises: AgentAPIError when agent failed to execute specified command.
        :raises: AgentInProgress when the command fails to execute as the agent
                 is presently executing the prior command.
        :returns: A dict containing command response from agent.
                  See :func:`get_commands_status` for a command result sample.
        """
        params = {'root_uuid': root_uuid,
                  'efi_system_part_uuid': efi_system_part_uuid,
                  'prep_boot_part_uuid': prep_boot_part_uuid,
                  'target_boot_mode': target_boot_mode
                  }

        # NOTE(TheJulia): This command explicitly sends a larger timeout
        # factor to the _command call such that the agent ramdisk has enough
        # time to perform its work.
        # TODO(TheJulia): We should likely split install_bootloader into many
        # commands at some point, even though that would not be backwards
        # compatible. We could at least begin to delineate the commands apart
        # over the next cycle or two so we don't need a command timeout
        # extension factor.
        try:
            return self._command(node=node,
                                 method='image.install_bootloader',
                                 params=params,
                                 poll=True)
        except exception.AgentAPIError:
            # NOTE(arne_wiebalck): If for software RAID and 'uefi' as the boot
            # mode, we find that the IPA does not yet support the additional
            # 'target_boot_mode' parameter, we need to fail. For 'bios' boot
            # mode on the other hand we can retry without the parameter,
            # since 'bios' is the default value the IPA will use.
            if target_boot_mode == 'uefi' and software_raid:
                LOG.error('Unable to pass UEFI boot mode to an out of date '
                          'agent ramdisk. Please contact the administrator '
                          'to update the ramdisk to contain an '
                          'ironic-python-agent version of at least 6.0.0.')
                raise
            else:
                params = {'root_uuid': root_uuid,
                          'efi_system_part_uuid': efi_system_part_uuid,
                          'prep_boot_part_uuid': prep_boot_part_uuid
                          }
                LOG.warning('Failed to install bootloader on first attempt '
                            'for node %(node)s. Falling back to older IPA '
                            'format.', {'node': node.uuid})
                return self._command(node=node,
                                     method='image.install_bootloader',
                                     params=params,
                                     poll=True)

    @METRICS.timer('AgentClient.get_clean_steps')
    def get_clean_steps(self, node, ports):
        """Get clean steps from agent.

        :param node: A node object.
        :param ports: Ports associated with the node.
        :raises: IronicException when failed to issue the request or there was
                 a malformed response from the agent.
        :raises: AgentAPIError when agent failed to execute specified command.
        :raises: AgentInProgress when the command fails to execute as the agent
                 is presently executing the prior command.
        :returns: A dict containing command response from agent.
            See :func:`get_commands_status` for a command result sample.
            The value of key command_result is in the form of:

            ::

              {
                'clean_steps': <a list of clean steps>,
                'hardware_manager_version': <manager version>
              }

        """
        params = {
            'node': node.as_dict(secure=True),
            'ports': [port.as_dict() for port in ports]
        }
        return self._command(node=node,
                             method='clean.get_clean_steps',
                             params=params,
                             wait=True)

    @METRICS.timer('AgentClient.execute_clean_step')
    def execute_clean_step(self, step, node, ports):
        """Execute specified clean step.

        :param step: A clean step dictionary to execute.
        :param node: A Node object.
        :param ports: Ports associated with the node.
        :raises: IronicException when failed to issue the request or there was
                 a malformed response from the agent.
        :raises: AgentAPIError when agent failed to execute specified command.
        :raises: AgentInProgress when the command fails to execute as the agent
                 is presently executing the prior command.
        :returns: A dict containing command response from agent.
            See :func:`get_commands_status` for a command result sample.
            The value of key command_result is in the form of:

            ::

              {
                'clean_result': <the result of execution, step specific>,
                'clean_step': <the clean step issued to agent>
              }

        """
        params = {
            'step': step,
            'node': node.as_dict(secure=True),
            'ports': [port.as_dict() for port in ports],
            'clean_version': node.driver_internal_info.get(
                'hardware_manager_version')
        }
        return self._command(node=node,
                             method='clean.execute_clean_step',
                             params=params)

    @METRICS.timer('AgentClient.get_deploy_steps')
    def get_deploy_steps(self, node, ports):
        """Get deploy steps from agent.

        :param node: A node object.
        :param ports: Ports associated with the node.
        :raises: IronicException when failed to issue the request or there was
                 a malformed response from the agent.
        :raises: AgentAPIError when agent failed to execute specified command.
        :returns: A dict containing command response from agent.
            See :func:`get_commands_status` for a command result sample.
            The value of key command_result is in the form of:

            ::

              {
                'deploy_steps': <a list of deploy steps>,
                'hardware_manager_version': <manager version>
              }

        """
        params = {
            'node': node.as_dict(secure=True),
            'ports': [port.as_dict() for port in ports]
        }
        return self._command(node=node,
                             method='deploy.get_deploy_steps',
                             params=params,
                             wait=True)

    @METRICS.timer('AgentClient.execute_deploy_step')
    def execute_deploy_step(self, step, node, ports):
        """Execute specified deploy step.

        :param step: A deploy step dictionary to execute.
        :param node: A Node object.
        :param ports: Ports associated with the node.
        :raises: IronicException when failed to issue the request or there was
                 a malformed response from the agent.
        :raises: AgentAPIError when agent failed to execute specified command.
        :raises: AgentInProgress when the command fails to execute as the agent
                 is presently executing the prior command.
        :returns: A dict containing command response from agent.
            See :func:`get_commands_status` for a command result sample.
            The value of key command_result is in the form of:

            ::

              {
                'deploy_result': <the result of execution, step specific>,
                'deploy_step': <the deploy step issued to agent>
              }

        """
        params = {
            'step': step,
            'node': node.as_dict(secure=True, mask_configdrive=False),
            'ports': [port.as_dict() for port in ports],
            'deploy_version': node.driver_internal_info.get(
                'hardware_manager_version')
        }
        return self._command(node=node,
                             method='deploy.execute_deploy_step',
                             params=params)

    @METRICS.timer('AgentClient.get_partition_uuids')
    def get_partition_uuids(self, node):
        """Get deploy steps from agent.

        :param node: A node object.
        :raises: IronicException when failed to issue the request or there was
                 a malformed response from the agent.
        :raises: AgentAPIError when agent failed to execute specified command.
        :raises: AgentInProgress when the command fails to execute as the agent
                 is presently executing the prior command.
        :returns: A dict containing command response from agent.

        """
        return self._command(node=node,
                             method='standby.get_partition_uuids',
                             params={},
                             wait=True)

    @METRICS.timer('AgentClient.power_off')
    def power_off(self, node):
        """Soft powers off the bare metal node by shutting down ramdisk OS.

        :param node: A Node object.
        :raises: IronicException when failed to issue the request or there was
                 a malformed response from the agent.
        :raises: AgentAPIError when agent failed to execute specified command.
        :raises: AgentInProgress when the command fails to execute as the agent
                 is presently executing the prior command.
        :returns: A dict containing command response from agent.
                  See :func:`get_commands_status` for a command result sample.
        """
        return self._command(node=node,
                             method='standby.power_off',
                             params={})

    @METRICS.timer('AgentClient.reboot')
    def reboot(self, node):
        """Soft reboots the bare metal node by shutting down ramdisk OS.

        :param node: A Node object.
        :raises: IronicException when failed to issue the request or there was
                 a malformed response from the agent.
        :raises: AgentAPIError when agent failed to execute specified command.
        :raises: AgentInProgress when the command fails to execute as the agent
                 is presently executing the prior command.
        :returns: A dict containing command response from agent.
                  See :func:`get_commands_status` for a command result sample.
        """
        return self._command(node=node,
                             method='standby.%s' % REBOOT_COMMAND,
                             params={})

    @METRICS.timer('AgentClient.sync')
    def sync(self, node):
        """Flush file system buffers forcing changed blocks to disk.

        :param node: A Node object.
        :raises: IronicException when failed to issue the request or there was
                 a malformed response from the agent.
        :raises: AgentAPIError when agent failed to execute specified command.
        :raises: AgentInProgress when the command fails to execute as the agent
                 is presently executing the prior command.
        :returns: A dict containing command response from agent.
                  See :func:`get_commands_status` for a command result sample.
        """
        return self._command(node=node,
                             method='standby.sync',
                             params={},
                             wait=True)

    @METRICS.timer('AgentClient.collect_system_logs')
    def collect_system_logs(self, node):
        """Collect and package diagnostic and support data from the ramdisk.

        :param node: A Node object.
        :raises: IronicException when failed to issue the request or there was
                 a malformed response from the agent.
        :raises: AgentAPIError when agent failed to execute specified command.
        :raises: AgentInProgress when the command fails to execute as the agent
                 is presently executing the prior command.
        :returns: A dict containing command response from agent.
                  See :func:`get_commands_status` for a command result sample.
        """
        return self._command(node=node,
                             method='log.collect_system_logs',
                             params={},
                             wait=True)

    @METRICS.timer('AgentClient.finalize_rescue')
    def finalize_rescue(self, node):
        """Instruct the ramdisk to finalize entering of rescue mode.

        :param node: A Node object.
        :raises: IronicException if rescue_password is missing, or when failed
                 to issue the request, or there was a malformed response from
                 the agent.
        :raises: AgentAPIError when agent failed to execute specified command.
        :raises: AgentInProgress when the command fails to execute as the agent
                 is presently executing the prior command.
        :raises: InstanceRescueFailure when the agent ramdisk is too old
                 to support transmission of the rescue password.
        :returns: A dict containing command response from agent.
                  See :func:`get_commands_status` for a command result sample.
        """
        rescue_pass = node.instance_info.get('hashed_rescue_password')
        # TODO(TheJulia): Remove fallback to use the fallback_rescue_password
        # in the Victoria cycle.
        fallback_rescue_pass = node.instance_info.get(
            'rescue_password')
        if not rescue_pass:
            raise exception.IronicException(_('Agent rescue requires '
                                              'rescue_password in '
                                              'instance_info'))
        params = {'rescue_password': rescue_pass,
                  'hashed': True}
        try:
            return self._command(node=node,
                                 method='rescue.finalize_rescue',
                                 params=params)
        except exception.AgentAPIError:
            if CONF.conductor.require_rescue_password_hashed:
                raise exception.InstanceRescueFailure(
                    _('Unable to rescue node due to an out of date agent '
                      'ramdisk. Please contact the administrator to update '
                      'the rescue ramdisk to contain an ironic-python-agent '
                      'version of at least 6.0.0.'))
            else:
                params = {'rescue_password': fallback_rescue_pass}
                return self._command(node=node,
                                     method='rescue.finalize_rescue',
                                     params=params)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Added portgroups table and altered ports

Revision ID: 5ea1b0d310e
Revises: 48d6c242bb9b
Create Date: 2015-06-30 14:14:26.972368

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '5ea1b0d310e'
down_revision = '48d6c242bb9b'


def upgrade():
    op.create_table('portgroups',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('uuid', sa.String(length=36), nullable=True),
                    sa.Column('name', sa.String(length=255), nullable=True),
                    sa.Column('node_id', sa.Integer(), nullable=True),
                    sa.Column('address', sa.String(length=18), nullable=True),
                    sa.Column('extra', sa.Text(), nullable=True),
                    sa.ForeignKeyConstraint(['node_id'], ['nodes.id'], ),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('uuid', name='uniq_portgroups0uuid'),
                    sa.UniqueConstraint('address',
                                        name='uniq_portgroups0address'),
                    sa.UniqueConstraint('name', name='uniq_portgroups0name'),
                    mysql_ENGINE='InnoDB',
                    mysql_DEFAULT_CHARSET='UTF8')
    op.add_column(u'ports', sa.Column('local_link_connection', sa.Text(),
                                      nullable=True))
    op.add_column(u'ports', sa.Column('portgroup_id', sa.Integer(),
                                      nullable=True))
    op.add_column(u'ports', sa.Column('pxe_enabled', sa.Boolean(),
                                      default=True))
    op.create_foreign_key('fk_portgroups_ports', 'ports', 'portgroups',
                          ['portgroup_id'], ['id'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils
from oslo_versionedobjects import base as object_base

from ironic.common import exception
from ironic.db import api as dbapi
from ironic.objects import base
from ironic.objects import fields as object_fields
from ironic.objects import node as node_obj


@base.IronicObjectRegistry.register
class Deployment(base.IronicObject, object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    VERSION = '1.0'

    dbapi = dbapi.get_instance()

    fields = {
        'uuid': object_fields.UUIDField(nullable=True),
        'node_uuid': object_fields.UUIDField(nullable=True),
        'image_checksum': object_fields.StringField(nullable=True),
        'image_ref': object_fields.StringField(nullable=True),
        'kernel_ref': object_fields.StringField(nullable=True),
        'ramdisk_ref': object_fields.StringField(nullable=True),
        'root_device': object_fields.FlexibleDictField(nullable=True),
        'root_gib': object_fields.IntegerField(nullable=True),
        'state': object_fields.StringField(nullable=True),
        'swap_mib': object_fields.IntegerField(nullable=True),
    }

    node_mapping = {
        'instance_uuid': 'uuid',
        'provision_state': 'state',
        'uuid': 'node_uuid',
    }

    instance_info_mapping = {
        'image_checksum': 'image_checksum',
        'image_source': 'image_ref',
        'kernel': 'kernel_ref',
        'ramdisk': 'ramdisk_ref',
        'root_device': 'root_device',
        'root_gb': 'root_gib',
        'swap_mb': 'swap_mib',
    }

    instance_info_mapping_rev = {v: k
                                 for k, v in instance_info_mapping.items()}

    assert (set(node_mapping.values()) | set(instance_info_mapping.values())
            == set(fields))

    def _convert_to_version(self, target_version,
                            remove_unavailable_fields=True):
        """Convert to the target version.

        Convert the object to the target version. The target version may be
        the same, older, or newer than the version of the object. This is
        used for DB interactions as well as for serialization/deserialization.

        :param target_version: the desired version of the object
        :param remove_unavailable_fields: True to remove fields that are
            unavailable in the target version; set this to True when
            (de)serializing. False to set the unavailable fields to appropriate
            values; set this to False for DB interactions.
        """

    @classmethod
    def _from_node_object(cls, context, node):
        """Convert a node into a virtual `Deployment` object."""
        result = cls(context)
        result._update_from_node_object(node)
        return result

    def _update_from_node_object(self, node):
        """Update the Deployment object from the node."""
        for src, dest in self.node_mapping.items():
            setattr(self, dest, getattr(node, src, None))
        for src, dest in self.instance_info_mapping.items():
            setattr(self, dest, node.instance_info.get(src))

    def _update_node_object(self, node):
        """Update the given node object with the changes here."""
        changes = self.obj_get_changes()
        try:
            new_instance_uuid = changes.pop('uuid')
        except KeyError:
            pass
        else:
            node.instance_uuid = new_instance_uuid

        changes.pop('node_uuid', None)
        instance_info = node.instance_info

        for field, value in changes.items():
            # NOTE(dtantsur): only instance_info fields can be updated here.
            try:
                dest = self.instance_info_mapping_rev[field]
            except KeyError:
                # NOTE(dtantsur): this should not happen because of API-level
                # validations, but checking just in case.
                raise exception.BadRequest('Field %s cannot be set or updated'
                                           % changes)
            instance_info[dest] = value

        node.instance_info = instance_info
        return node

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def get_by_uuid(cls, context, uuid):
        """Find a deployment by its UUID.

        :param cls: the :class:`Deployment`
        :param context: Security context
        :param uuid: The UUID of a deployment.
        :returns: An :class:`Deployment` object.
        :raises: InstanceNotFound

        """
        node = node_obj.Node.get_by_instance_uuid(context, uuid)
        return cls._from_node_object(context, node)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def get_by_node_uuid(cls, context, node_uuid):
        """Find a deployment based by its node's UUID.

        :param cls: the :class:`Deployment`
        :param context: Security context
        :param node_uuid: The UUID of a corresponding node.
        :returns: An :class:`Deployment` object.
        :raises: NodeNotFound

        """
        node = node_obj.Node.get_by_uuid(context, node_uuid)
        return cls._from_node_object(context, node)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def list(cls, context, filters=None, limit=None, marker=None,
             sort_key=None, sort_dir=None):
        """Return a list of Deployment objects.

        :param cls: the :class:`Deployment`
        :param context: Security context.
        :param filters: Filters to apply.
        :param limit: Maximum number of resources to return in a single result.
        :param marker: Pagination marker for large data sets.
        :param sort_key: Column to sort results by.
        :param sort_dir: Direction to sort. "asc" or "desc".
        :returns: A list of :class:`Deployment` object.
        :raises: InvalidParameterValue

        """
        nodes = node_obj.Node.list(context, filters=filters, limit=limit,
                                   marker=marker, sort_key=sort_key,
                                   sort_dir=sort_dir)
        return [cls._from_node_object(context, node) for node in nodes]

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable
    def create(self, context=None, node=None):
        """Create a Deployment.

        Updates the corresponding node under the hood.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Deployment(context)
        :param node: Node object for deployment.
        :raises: InstanceAssociated, NodeAssociated, NodeNotFound

        """
        if node is None:
            node = node_obj.Node.get_by_uuid(self._context, self.node_uuid)
        elif 'node_uuid' in self and self.node_uuid:
            # NOTE(dtantsur): this is only possible if a bug happens on
            # a higher level.
            assert self.node_uuid == node.uuid

        if 'uuid' not in self or not self.uuid:
            self.uuid = uuidutils.generate_uuid()
        node.instance_uuid = self.uuid
        self._update_node_object(node)
        node.save()
        self._update_from_node_object(node)
        self.obj_reset_changes()

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable
    def destroy(self, context=None, node=None):
        """Delete the Deployment.

        Updates the corresponding node under the hood.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Node(context)
        :param node: Node object for deployment.
        """
        if node is None:
            node = node_obj.Node.get_by_uuid(self._context, self.node_uuid)
        else:
            assert node.uuid == self.node_uuid
        node.instance_uuid = None
        node.instance_info = {}
        node.save()
        self._update_from_node_object(node)
        self.obj_reset_changes()

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable
    def refresh(self, context=None):
        """Refresh the object by re-fetching from the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Node(context)
        """
        current = self.get_by_uuid(self._context, self.uuid)
        self.obj_refresh(current)
        self.obj_reset_changes()



# Copyright 2018 DMTF. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironic_lib import metrics_utils
from oslo_log import log
from oslo_utils import importutils

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import states
from ironic.conductor import task_manager
from ironic.conductor import utils as manager_utils
from ironic.drivers import base
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules.redfish import utils as redfish_utils
from ironic import objects

LOG = log.getLogger(__name__)

METRICS = metrics_utils.get_metrics_logger(__name__)

sushy = importutils.try_import('sushy')

registry_fields = ('attribute_type', 'allowable_values', 'lower_bound',
                   'max_length', 'min_length', 'read_only',
                   'reset_required', 'unique', 'upper_bound')


class RedfishBIOS(base.BIOSInterface):

    _APPLY_CONFIGURATION_ARGSINFO = {
        'settings': {
            'description': (
                'A list of BIOS settings to be applied'
            ),
            'required': True
        }
    }

    def __init__(self):
        super(RedfishBIOS, self).__init__()
        if sushy is None:
            raise exception.DriverLoadError(
                driver='redfish',
                reason=_("Unable to import the sushy library"))

    def _parse_allowable_values(self, allowable_values):
        """Convert the BIOS registry allowable_value list to expected strings

        :param allowable_values: list of dicts of valid values for enumeration
        :returns: list containing only allowable value names
        """

        # Get name from ValueName if it exists, otherwise use DisplayValueName
        new_list = []
        for dic in allowable_values:
            for key in dic:
                if key == 'ValueName' or key == 'DisplayValueName':
                    new_list.append(dic[key])
                    break

        return new_list

    def cache_bios_settings(self, task):
        """Store or update the current BIOS settings for the node.

        Get the current BIOS settings and store them in the bios_settings
        database table.

        :param task: a TaskManager instance containing the node to act on.
        :raises: RedfishConnectionError when it fails to connect to Redfish
        :raises: RedfishError on an error from the Sushy library
        :raises: UnsupportedDriverExtension if the system does not support BIOS
            settings
        """

        node_id = task.node.id
        system = redfish_utils.get_system(task.node)
        try:
            attributes = system.bios.attributes
        except sushy.exceptions.MissingAttributeError:
            error_msg = _('Cannot fetch BIOS attributes for node %s, '
                          'BIOS settings are not supported.') % task.node.uuid
            LOG.error(error_msg)
            raise exception.UnsupportedDriverExtension(error_msg)

        settings = []
        # Convert Redfish BIOS attributes to Ironic BIOS settings
        if attributes:
            settings = [{'name': k, 'value': v}
                        for k, v in attributes.items()]

        # Get the BIOS Registry
        registry_attributes = []
        try:
            bios_registry = system.bios.get_attribute_registry()

            if bios_registry:
                registry_attributes = bios_registry.registry_entries.attributes

        except Exception as e:
            LOG.info('Cannot get BIOS Registry attributes for node %(node)s, '
                     'Error %(exc)s.', {'node': task.node.uuid, 'exc': e})

        # TODO(bfournier): use a common list for registry field names
        # e.g. registry_fields = objects.BIOSSetting.registry_fields

        # The BIOS registry will contain more entries than the BIOS settings
        # Find the registry entry matching the setting name and get the fields
        if registry_attributes:
            for setting in settings:
                reg = next((r for r in registry_attributes
                            if r.name == setting['name']), None)
                fields = [attr for attr in dir(reg)
                          if not attr.startswith("_")]
                settable_keys = [f for f in fields if f in registry_fields]
                # Set registry fields to current values
                for k in settable_keys:
                    setting[k] = getattr(reg, k, None)
                    if k == "allowable_values" and isinstance(setting[k],
                                                              list):
                        setting[k] = self._parse_allowable_values(setting[k])

        LOG.debug('Cache BIOS settings for node %(node_uuid)s',
                  {'node_uuid': task.node.uuid})

        create_list, update_list, delete_list, nochange_list = (
            objects.BIOSSettingList.sync_node_setting(
                task.context, node_id, settings))

        if create_list:
            objects.BIOSSettingList.create(
                task.context, node_id, create_list)
        if update_list:
            objects.BIOSSettingList.save(
                task.context, node_id, update_list)
        if delete_list:
            delete_names = [d['name'] for d in delete_list]
            objects.BIOSSettingList.delete(
                task.context, node_id, delete_names)

    @base.clean_step(priority=0)
    @base.deploy_step(priority=0)
    @base.cache_bios_settings
    def factory_reset(self, task):
        """Reset the BIOS settings of the node to the factory default.

        :param task: a TaskManager instance containing the node to act on.
        :raises: RedfishConnectionError when it fails to connect to Redfish
        :raises: RedfishError on an error from the Sushy library
        """
        system = redfish_utils.get_system(task.node)
        try:
            bios = system.bios
        except sushy.exceptions.MissingAttributeError:
            error_msg = (_('Redfish BIOS factory reset failed for node '
                           '%s, because BIOS settings are not supported.') %
                         task.node.uuid)
            LOG.error(error_msg)
            raise exception.RedfishError(error=error_msg)

        node = task.node
        info = node.driver_internal_info
        reboot_requested = info.get('post_factory_reset_reboot_requested')
        if not reboot_requested:
            LOG.debug('Factory reset BIOS configuration for node %(node)s',
                      {'node': node.uuid})
            try:
                bios.reset_bios()
            except sushy.exceptions.SushyError as e:
                error_msg = (_('Redfish BIOS factory reset failed for node '
                               '%(node)s. Error: %(error)s') %
                             {'node': node.uuid, 'error': e})
                LOG.error(error_msg)
                raise exception.RedfishError(error=error_msg)

            self.post_reset(task)
            self._set_reboot(task)
            return deploy_utils.get_async_step_return_state(task.node)
        else:
            current_attrs = bios.attributes
            LOG.debug('Post factory reset, BIOS configuration for node '
                      '%(node_uuid)s: %(attrs)r',
                      {'node_uuid': node.uuid, 'attrs': current_attrs})
            self._clear_reboot_requested(task)

    @base.clean_step(priority=0, argsinfo=_APPLY_CONFIGURATION_ARGSINFO)
    @base.deploy_step(priority=0, argsinfo=_APPLY_CONFIGURATION_ARGSINFO)
    @base.cache_bios_settings
    def apply_configuration(self, task, settings):
        """Apply the BIOS settings to the node.

        :param task: a TaskManager instance containing the node to act on.
        :param settings: a list of BIOS settings to be updated.
        :raises: RedfishConnectionError when it fails to connect to Redfish
        :raises: RedfishError on an error from the Sushy library
        """

        system = redfish_utils.get_system(task.node)
        try:
            bios = system.bios
        except sushy.exceptions.MissingAttributeError:
            error_msg = (_('Redfish BIOS factory reset failed for node '
                           '%s, because BIOS settings are not supported.') %
                         task.node.uuid)
            LOG.error(error_msg)
            raise exception.RedfishError(error=error_msg)

        # Convert Ironic BIOS settings to Redfish BIOS attributes
        attributes = {s['name']: s['value'] for s in settings}

        info = task.node.driver_internal_info
        reboot_requested = info.get('post_config_reboot_requested')

        if not reboot_requested:
            # Step 1: Apply settings and issue a reboot
            LOG.debug('Apply BIOS configuration for node %(node_uuid)s: '
                      '%(settings)r', {'node_uuid': task.node.uuid,
                                       'settings': settings})

            if bios.supported_apply_times and (
                    sushy.APPLY_TIME_ON_RESET in bios.supported_apply_times):
                apply_time = sushy.APPLY_TIME_ON_RESET
            else:
                apply_time = None

            try:
                bios.set_attributes(attributes, apply_time=apply_time)
            except sushy.exceptions.SushyError as e:
                error_msg = (_('Redfish BIOS apply configuration failed for '
                               'node %(node)s. Error: %(error)s') %
                             {'node': task.node.uuid, 'error': e})
                LOG.error(error_msg)
                raise exception.RedfishError(error=error_msg)

            self.post_configuration(task, settings)
            self._set_reboot_requested(task, attributes)
            return deploy_utils.get_async_step_return_state(task.node)
        else:
            # Step 2: Verify requested BIOS settings applied
            requested_attrs = info.get('requested_bios_attrs')
            current_attrs = bios.attributes
            LOG.debug('Verify BIOS configuration for node %(node_uuid)s: '
                      '%(attrs)r', {'node_uuid': task.node.uuid,
                                    'attrs': requested_attrs})
            self._clear_reboot_requested(task)
            self._check_bios_attrs(task, current_attrs, requested_attrs)

    def post_reset(self, task):
        """Perform post reset action to apply the BIOS factory reset.

        Extension point to allow vendor implementations to extend this class
        and override this method to perform a custom action to apply the BIOS
        factory reset to the Redfish service. The default implementation
        performs a reboot.

        :param task: a TaskManager instance containing the node to act on.
        """
        deploy_utils.prepare_agent_boot(task)
        self._reboot(task)

    def post_configuration(self, task, settings):
        """Perform post configuration action to store the BIOS settings.

        Extension point to allow vendor implementations to extend this class
        and override this method to perform a custom action to write the BIOS
        settings to the Redfish service. The default implementation performs
        a reboot.

        :param task: a TaskManager instance containing the node to act on.
        :param settings: a list of BIOS settings to be updated.
        """
        deploy_utils.prepare_agent_boot(task)
        self._reboot(task)

    def get_properties(self):
        """Return the properties of the interface.

        :returns: dictionary of <property name>:<property description> entries.
        """
        return redfish_utils.COMMON_PROPERTIES.copy()

    def validate(self, task):
        """Validates the driver information needed by the redfish driver.

        :param task: a TaskManager instance containing the node to act on.
        :raises: InvalidParameterValue on malformed parameter(s)
        :raises: MissingParameterValue on missing parameter(s)
        """
        redfish_utils.parse_driver_info(task.node)

    def _check_bios_attrs(self, task, current_attrs, requested_attrs):
        """Checks that the requested BIOS settings were applied to the service.

        :param task: a TaskManager instance containing the node to act on.
        :param current_attrs: the current BIOS attributes from the system.
        :param requested_attrs: the requested BIOS attributes to update.
        """

        attrs_not_updated = {}
        for attr in requested_attrs:
            if requested_attrs[attr] != current_attrs.get(attr):
                attrs_not_updated[attr] = requested_attrs[attr]

        if attrs_not_updated:
            LOG.debug('BIOS settings %(attrs)s for node %(node_uuid)s '
                      'not updated.', {'attrs': attrs_not_updated,
                                       'node_uuid': task.node.uuid})
            self._set_step_failed(task, attrs_not_updated)
        else:
            LOG.debug('Verification of BIOS settings for node %(node_uuid)s '
                      'successful.', {'node_uuid': task.node.uuid})

    @task_manager.require_exclusive_lock
    def _reboot(self, task):
        """Reboot the target Redfish service.

        :param task: a TaskManager instance containing the node to act on.
        :raises: InvalidParameterValue when the wrong state is specified
             or the wrong driver info is specified.
        :raises: RedfishError on an error from the Sushy library
        """
        manager_utils.node_power_action(task, states.REBOOT)

    def _set_reboot(self, task):
        """Set driver_internal_info flags for deployment or cleaning reboot.

        :param task: a TaskManager instance containing the node to act on.
        """
        task.node.set_driver_internal_info(
            'post_factory_reset_reboot_requested', True)
        task.node.save()
        deploy_utils.set_async_step_flags(task.node, reboot=True,
                                          skip_current_step=False)

    def _set_reboot_requested(self, task, attributes):
        """Set driver_internal_info flags for reboot requested.

        :param task: a TaskManager instance containing the node to act on.
        :param attributes: the requested BIOS attributes to update.
        """
        task.node.set_driver_internal_info('post_config_reboot_requested',
                                           True)
        task.node.set_driver_internal_info('requested_bios_attrs', attributes)
        task.node.save()
        deploy_utils.set_async_step_flags(task.node, reboot=True,
                                          skip_current_step=False)

    def _clear_reboot_requested(self, task):
        """Clear driver_internal_info flags after reboot completed.

        :param task: a TaskManager instance containing the node to act on.
        """
        node = task.node
        node.del_driver_internal_info('post_config_reboot_requested')
        node.del_driver_internal_info('post_factory_reset_reboot_requested')
        node.del_driver_internal_info('requested_bios_attrs')
        node.save()

    def _set_step_failed(self, task, attrs_not_updated):
        """Fail the cleaning or deployment step and log the error.

        :param task: a TaskManager instance containing the node to act on.
        :param attrs_not_updated: the BIOS attributes that were not updated.
        """
        error_msg = (_('Redfish BIOS apply_configuration step failed for node '
                       '%(node)s. Attributes %(attrs)s are not updated.') %
                     {'node': task.node.uuid, 'attrs': attrs_not_updated})
        last_error = (_('Redfish BIOS apply_configuration step failed. '
                        'Attributes %(attrs)s are not updated.') %
                      {'attrs': attrs_not_updated})
        if task.node.provision_state in [states.CLEANING, states.CLEANWAIT]:
            manager_utils.cleaning_error_handler(task, last_error)
        if task.node.provision_state in [states.DEPLOYING, states.DEPLOYWAIT]:
            manager_utils.deploying_error_handler(task, error_msg, last_error)



# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for manipulating Ports via the DB API"""

from oslo_utils import uuidutils

from ironic.common import exception
from ironic.tests.unit.db import base
from ironic.tests.unit.db import utils as db_utils


class DbPortTestCase(base.DbTestCase):

    def setUp(self):
        # This method creates a port for every test and
        # replaces a test for creating a port.
        super(DbPortTestCase, self).setUp()
        self.node = db_utils.create_test_node(owner='12345',
                                              lessee='54321')
        self.portgroup = db_utils.create_test_portgroup(node_id=self.node.id)
        self.port = db_utils.create_test_port(node_id=self.node.id,
                                              portgroup_id=self.portgroup.id,
                                              name='port-name')

    def test_get_port_by_id(self):
        res = self.dbapi.get_port_by_id(self.port.id)
        self.assertEqual(self.port.address, res.address)

    def test_get_port_by_uuid(self):
        res = self.dbapi.get_port_by_uuid(self.port.uuid)
        self.assertEqual(self.port.id, res.id)

    def test_get_port_by_address(self):
        res = self.dbapi.get_port_by_address(self.port.address)
        self.assertEqual(self.port.id, res.id)

    def test_get_port_by_address_filter_by_owner(self):
        res = self.dbapi.get_port_by_address(self.port.address,
                                             owner=self.node.owner)
        self.assertEqual(self.port.id, res.id)

    def test_get_port_by_address_filter_by_owner_no_match(self):
        self.assertRaises(exception.PortNotFound,
                          self.dbapi.get_port_by_address,
                          self.port.address,
                          owner='54321')

    def test_get_port_by_address_filter_by_project(self):
        res = self.dbapi.get_port_by_address(self.port.address,
                                             project=self.node.lessee)
        self.assertEqual(self.port.id, res.id)

    def test_get_port_by_address_filter_by_project_no_match(self):
        self.assertRaises(exception.PortNotFound,
                          self.dbapi.get_port_by_address,
                          self.port.address,
                          project='55555')

    def test_get_port_by_name(self):
        res = self.dbapi.get_port_by_name(self.port.name)
        self.assertEqual(self.port.id, res.id)

    def test_get_port_list(self):
        uuids = []
        for i in range(1, 6):
            port = db_utils.create_test_port(uuid=uuidutils.generate_uuid(),
                                             node_id=self.node.id,
                                             address='52:54:00:cf:2d:4%s' % i)
            uuids.append(str(port.uuid))
        # Also add the uuid for the port created in setUp()
        uuids.append(str(self.port.uuid))
        res = self.dbapi.get_port_list()
        res_uuids = [r.uuid for r in res]
        self.assertCountEqual(uuids, res_uuids)

    def test_get_port_list_sorted(self):
        uuids = []
        for i in range(1, 6):
            port = db_utils.create_test_port(uuid=uuidutils.generate_uuid(),
                                             node_id=self.node.id,
                                             address='52:54:00:cf:2d:4%s' % i)
            uuids.append(str(port.uuid))
        # Also add the uuid for the port created in setUp()
        uuids.append(str(self.port.uuid))
        res = self.dbapi.get_port_list(sort_key='uuid')
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids), res_uuids)

        self.assertRaises(exception.InvalidParameterValue,
                          self.dbapi.get_port_list, sort_key='foo')

    def test_get_port_list_filter_by_node_owner(self):
        another_node = db_utils.create_test_node(
            uuid=uuidutils.generate_uuid())
        uuids = []
        for i in range(1, 3):
            port = db_utils.create_test_port(uuid=uuidutils.generate_uuid(),
                                             node_id=another_node.id,
                                             address='52:54:00:cf:2d:4%s' % i)
        for i in range(4, 6):
            port = db_utils.create_test_port(uuid=uuidutils.generate_uuid(),
                                             node_id=self.node.id,
                                             address='52:54:00:cf:2d:4%s' % i)
            uuids.append(str(port.uuid))
        # Also add the uuid for the port created in setUp()
        uuids.append(str(self.port.uuid))
        res = self.dbapi.get_port_list(owner=self.node.owner)
        res_uuids = [r.uuid for r in res]
        self.assertCountEqual(uuids, res_uuids)

    def test_get_port_list_filter_by_node_project(self):
        another_node = db_utils.create_test_node(
            uuid=uuidutils.generate_uuid())
        lessee_node = db_utils.create_test_node(uuid=uuidutils.generate_uuid(),
                                                lessee=self.node.owner)

        uuids = []
        for i in range(1, 3):
            port = db_utils.create_test_port(uuid=uuidutils.generate_uuid(),
                                             node_id=lessee_node.id,
                                             address='52:54:00:cf:2d:4%s' % i)
            uuids.append(str(port.uuid))
        for i in range(4, 6):
            port = db_utils.create_test_port(uuid=uuidutils.generate_uuid(),
                                             node_id=another_node.id,
                                             address='52:54:00:cf:2d:4%s' % i)
        for i in range(7, 9):
            port = db_utils.create_test_port(uuid=uuidutils.generate_uuid(),
                                             node_id=self.node.id,
                                             address='52:54:00:cf:2d:4%s' % i)
            uuids.append(str(port.uuid))
        # Also add the uuid for the port created in setUp()
        uuids.append(str(self.port.uuid))
        res = self.dbapi.get_port_list(project=self.node.owner)
        res_uuids = [r.uuid for r in res]
        self.assertCountEqual(uuids, res_uuids)

    def test_get_ports_by_node_id(self):
        res = self.dbapi.get_ports_by_node_id(self.node.id)
        self.assertEqual(self.port.address, res[0].address)

    def test_get_ports_by_node_id_filter_by_node_owner(self):
        res = self.dbapi.get_ports_by_node_id(self.node.id,
                                              owner=self.node.owner)
        self.assertEqual(self.port.address, res[0].address)

    def test_get_ports_by_node_id_filter_by_node_owner_no_match(self):
        res = self.dbapi.get_ports_by_node_id(self.node.id,
                                              owner='54321')
        self.assertEqual([], res)

    def test_get_ports_by_node_id_filter_by_node_project(self):
        res = self.dbapi.get_ports_by_node_id(self.node.id,
                                              project=self.node.lessee)
        self.assertEqual(self.port.address, res[0].address)

    def test_get_ports_by_node_id_filter_by_node_project_no_match(self):
        res = self.dbapi.get_ports_by_node_id(self.node.id,
                                              owner='11111')
        self.assertEqual([], res)

    def test_get_ports_by_node_id_that_does_not_exist(self):
        self.assertEqual([], self.dbapi.get_ports_by_node_id(99))

    def test_get_ports_by_portgroup_id(self):
        res = self.dbapi.get_ports_by_portgroup_id(self.portgroup.id)
        self.assertEqual(self.port.address, res[0].address)

    def test_get_ports_by_portgroup_id_filter_by_node_owner(self):
        res = self.dbapi.get_ports_by_portgroup_id(self.portgroup.id,
                                                   owner=self.node.owner)
        self.assertEqual(self.port.address, res[0].address)

    def test_get_ports_by_portgroup_id_filter_by_node_owner_no_match(self):
        res = self.dbapi.get_ports_by_portgroup_id(self.portgroup.id,
                                                   owner='54321')
        self.assertEqual([], res)

    def test_get_ports_by_portgroup_id_filter_by_node_project(self):
        res = self.dbapi.get_ports_by_portgroup_id(self.portgroup.id,
                                                   project=self.node.lessee)
        self.assertEqual(self.port.address, res[0].address)

    def test_get_ports_by_portgroup_id_filter_by_node_project_no_match(self):
        res = self.dbapi.get_ports_by_portgroup_id(self.portgroup.id,
                                                   project='11111')
        self.assertEqual([], res)

    def test_get_ports_by_portgroup_id_that_does_not_exist(self):
        self.assertEqual([], self.dbapi.get_ports_by_portgroup_id(99))

    def test_destroy_port(self):
        self.dbapi.destroy_port(self.port.id)
        self.assertRaises(exception.PortNotFound,
                          self.dbapi.destroy_port, self.port.id)

    def test_update_port(self):
        old_address = self.port.address
        new_address = 'ff.ee.dd.cc.bb.aa'

        self.assertNotEqual(old_address, new_address)

        res = self.dbapi.update_port(self.port.id, {'address': new_address})
        self.assertEqual(new_address, res.address)

    def test_update_port_uuid(self):
        self.assertRaises(exception.InvalidParameterValue,
                          self.dbapi.update_port, self.port.id,
                          {'uuid': ''})

    def test_update_port_duplicated_address(self):
        address1 = self.port.address
        address2 = 'aa-bb-cc-11-22-33'
        port2 = db_utils.create_test_port(uuid=uuidutils.generate_uuid(),
                                          node_id=self.node.id,
                                          address=address2)
        self.assertRaises(exception.MACAlreadyExists,
                          self.dbapi.update_port, port2.id,
                          {'address': address1})

    def test_create_port_duplicated_address(self):
        self.assertRaises(exception.MACAlreadyExists,
                          db_utils.create_test_port,
                          uuid=uuidutils.generate_uuid(),
                          node_id=self.node.id,
                          address=self.port.address)

    def test_create_port_duplicated_uuid(self):
        self.assertRaises(exception.PortAlreadyExists,
                          db_utils.create_test_port,
                          uuid=self.port.uuid,
                          node_id=self.node.id,
                          address='aa-bb-cc-33-11-22')
